import datetime
import enum
import re
from dataclasses import dataclass

from sqlalchemy import DateTime, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db
from opentakserver.functions import iso8601_string_from_datetime
from opentakserver.models.GroupUser import GroupUser


class GroupTypeEnum(str, enum.Enum):
    SYSTEM = "SYSTEM"
    LDAP = "LDAP"


class GroupDirectionEnum(str, enum.Enum):
    IN = "IN"  # Write
    OUT = "OUT"  # Read


@dataclass
class Group(db.Model):
    __tablename__ = "groups"

    IN = "IN"
    OUT = "OUT"
    SYSTEM = "SYSTEM"
    LDAP = "LDAP"

    def __init__(self):
        super().__init__()
        self.bitpos = self.get_next_bitpos()

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    name: Mapped[str] = mapped_column(String(255), unique=True)
    distinguishedName: Mapped[str] = mapped_column(String(255), nullable=True)
    created: Mapped[datetime] = mapped_column(
        DateTime, default=datetime.datetime.now(datetime.timezone.utc)
    )
    type: Mapped[str] = mapped_column(String(255))
    bitpos: Mapped[int] = mapped_column(Integer)
    description: Mapped[str] = mapped_column(String(255), nullable=True)
    users = relationship(
        "User",
        secondary="groups_users",
        viewonly=True,
        back_populates="groups",
        cascade="all, delete",
    )
    missions = relationship(
        "Mission",
        secondary="groups_missions",
        viewonly=True,
        back_populates="groups",
        cascade="all, delete",
    )

    def get_next_bitpos(self) -> int:
        # the __ANON__ group is always 2 so default to 3 here
        bitpos = 3
        max_bitpos = db.session.execute(
            db.session.query(Group).order_by(Group.bitpos.desc()).limit(1)
        ).first()
        if max_bitpos:
            bitpos = max_bitpos[0].bitpos + 1

        return bitpos

    def set_bitpos(self, new_bitpos: int = None):
        if new_bitpos:
            self.bitpos = new_bitpos
        else:
            self.bitpos = self.get_next_bitpos()

    def serialize(self):
        return {
            "name": self.name,
            "distinguishedName": self.distinguishedName,
            "created": self.created,
            "type": self.type,
            "bitpos": self.bitpos,
            "description": self.description,
        }

    def to_json(self):
        return_value = self.serialize()
        return_value["id"] = self.id
        return_value["bitpos"] = "{0:b}".format(self.bitpos)
        return_value["users"] = [user.serialize() for user in self.users]
        return_value["missions"] = [mission.serialize() for mission in self.missions]
        return return_value

    def to_marti_json_in(self):
        # Remove the _READ and _WRITE suffixes when using LDAP for groups
        remove_read = re.compile(re.escape("_read"), re.IGNORECASE)
        remove_write = re.compile(re.escape("_write"), re.IGNORECASE)
        group_name = remove_read.sub("", self.name)
        group_name = remove_write.sub("", group_name)

        return {
            "name": group_name,
            "direction": Group.IN,
            "created": iso8601_string_from_datetime(
                self.created or datetime.datetime.now(tz=datetime.timezone.utc)
            ).split("T")[0],
            "type": self.type,
            "bitpos": self.bitpos,
            "active": True,
            "description": self.description or "",
        }

    def to_marti_json_out(self):
        return_value = self.to_marti_json_in()
        return_value["direction"] = Group.OUT
        return return_value



import traceback

from flask import Blueprint, jsonify, request
from flask_babel import gettext
from flask_security import auth_required, roles_required
from sqlalchemy import update
from sqlalchemy.exc import IntegrityError

from opentakserver.blueprints.ots_api.api import paginate, search
from opentakserver.extensions import db, logger
from opentakserver.forms.device_profile_form import DeviceProfileForm
from opentakserver.models.DeviceProfiles import DeviceProfiles

device_profile_api_blueprint = Blueprint("device_profile_api_blueprint", __name__)


@device_profile_api_blueprint.route("/api/profiles")
@roles_required("administrator")
def get_device_profiles():
    query = db.session.query(DeviceProfiles)
    query = search(query, DeviceProfiles, "preference_key")
    query = search(query, DeviceProfiles, "preference_value")
    query = search(query, DeviceProfiles, "tool")
    return paginate(query)


@device_profile_api_blueprint.route("/api/profiles", methods=["POST"])
@auth_required()
@roles_required("administrator")
def add_device_profile():
    form = DeviceProfileForm()
    if not form.validate():
        return jsonify({"success": False, "errors": form.errors}), 400

    device_profile = DeviceProfiles()
    device_profile.from_wtf(form)

    try:
        db.session.add(device_profile)
        db.session.commit()
    except IntegrityError:
        db.session.rollback()
        db.session.execute(
            update(DeviceProfiles)
            .where(DeviceProfiles.preference_key == device_profile.preference_key)
            .where(DeviceProfiles.eud_uid == device_profile.eud_uid)
            .values(**device_profile.serialize())
        )
        db.session.commit()

    return jsonify({"success": True})


@device_profile_api_blueprint.route("/api/profiles", methods=["DELETE"])
@roles_required("administrator")
def delete_device_profile():
    preference_key = request.args.get("preference_key")
    eud_uid = request.args.get("eud_uid")
    if not preference_key:
        return (
            jsonify({"success": False, "error": gettext("Please specify the preference_key")}),
            400,
        )
    try:
        query = db.session.query(DeviceProfiles)
        query = search(query, DeviceProfiles, "preference_key")
        query = search(query, DeviceProfiles, "eud_uid")
        preference = db.session.execute(query).first()
        if not preference:
            return (
                jsonify(
                    {
                        "success": False,
                        "error": gettext(
                            "Unknown preference_key: %(preference_key)s",
                            preference_key=preference_key,
                        ),
                    }
                ),
                404,
            )

        db.session.delete(preference[0])
        db.session.commit()
        return jsonify({"success": True})
    except BaseException as e:
        logger.error(f"Failed to delete device profile: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({"success": False, "error": str(e)}), 400



"""Fixes for the packages table, added group_eud table

Revision ID: f107b45529ba
Revises: 7e9f5d2c193d
Create Date: 2024-10-28 14:36:54.189219

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f107b45529ba"
down_revision = "7e9f5d2c193d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "groups_euds",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("group_id", sa.Integer(), nullable=False),
        sa.Column("direction", sa.String(length=255), nullable=False),
        sa.Column("active", sa.Boolean(), nullable=False),
        sa.Column("description", sa.String(length=255), nullable=True),
        sa.Column("eud_uid", sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(["eud_uid"], ["euds.uid"], name="group_eud"),
        sa.ForeignKeyConstraint(["group_id"], ["groups.id"], name="group"),
        sa.PrimaryKeyConstraint("id"),
    )

    with op.batch_alter_table("groups", schema=None) as batch_op:
        batch_op.drop_constraint("eud", type_="foreignkey")
        batch_op.drop_column("description")
        batch_op.drop_column("active")
        batch_op.drop_column("direction")
        batch_op.drop_column("eud_uid")

    with op.batch_alter_table("packages", schema=None) as batch_op:
        batch_op.alter_column(
            "description",
            existing_type=sa.Integer(),
            type_=sa.String(length=255),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "apk_hash",
            existing_type=sa.Integer(),
            type_=sa.String(length=255),
            existing_nullable=True,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("packages", schema=None) as batch_op:
        batch_op.alter_column(
            "apk_hash",
            existing_type=sa.String(length=255),
            type_=sa.Integer(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "description",
            existing_type=sa.String(length=255),
            type_=sa.Integer(),
            existing_nullable=True,
        )

    with op.batch_alter_table("groups", schema=None) as batch_op:
        batch_op.add_column(sa.Column("eud_uid", sa.String(length=255), nullable=False))
        batch_op.add_column(sa.Column("direction", sa.String(length=255), nullable=False))
        batch_op.add_column(sa.Column("active", sa.Boolean(), autoincrement=False, nullable=False))
        batch_op.add_column(sa.Column("description", sa.String(length=255), nullable=True))
        batch_op.create_foreign_key("eud", "euds", ["eud_uid"], ["uid"])

    op.drop_table("groups_euds")
    # ### end Alembic commands ###



from datetime import datetime

from sqlalchemy import JSON, TEXT, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db
from opentakserver.functions import iso8601_string_from_datetime


class CoT(db.Model):
    __tablename__ = "cot"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    how: Mapped[str] = mapped_column(String(255), nullable=True)
    type: Mapped[str] = mapped_column(String(255), nullable=True)
    uid: Mapped[str] = mapped_column(String(255), nullable=True)
    sender_callsign: Mapped[str] = mapped_column(String(255), nullable=True)
    sender_device_name: Mapped[str] = mapped_column(String(255), nullable=True)
    sender_uid: Mapped[str] = mapped_column(
        String(255), ForeignKey("euds.uid", ondelete="CASCADE"), nullable=True
    )
    recipients: Mapped[JSON] = mapped_column(JSON, nullable=True)
    timestamp: Mapped[datetime] = mapped_column(DateTime)
    start: Mapped[datetime] = mapped_column(DateTime)
    stale: Mapped[datetime] = mapped_column(DateTime)
    xml: Mapped[str] = mapped_column(TEXT)
    mission_name: Mapped[str] = mapped_column(
        String(255), ForeignKey("missions.name"), nullable=True
    )
    eud = relationship("EUD", back_populates="cots", uselist=False)
    alert = relationship(
        "Alert", cascade="all, delete, delete-orphan", back_populates="cot", uselist=False
    )
    point = relationship(
        "Point", cascade="all, delete, delete-orphan", back_populates="cot", uselist=False
    )
    casevac = relationship(
        "CasEvac", cascade="all, delete, delete-orphan", back_populates="cot", uselist=False
    )
    video = relationship(
        "VideoStream", back_populates="cot", cascade="all, delete, delete-orphan", uselist=False
    )
    geochat = relationship(
        "GeoChat", back_populates="cot", cascade="all, delete, delete-orphan", uselist=False
    )
    marker = relationship(
        "Marker", back_populates="cot", cascade="all, delete, delete-orphan", uselist=False
    )
    rb_line = relationship("RBLine", cascade="all, delete, delete-orphan", back_populates="cot")
    mission = relationship("Mission", back_populates="cots")

    def serialize(self):
        return {
            "how": self.how,
            "type": self.type,
            "uid": self.uid,
            "sender_callsign": self.sender_callsign,
            "sender_uid": self.sender_uid,
            "recipients": self.recipients,
            "timestamp": self.timestamp,
            "start": self.start,
            "stale": self.stale,
            "xml": self.xml,
        }

    def to_json(self):
        return {
            "how": self.how,
            "type": self.type,
            "uid": self.uid,
            "sender_callsign": self.sender_callsign,
            "sender_uid": self.sender_uid,
            "recipients": self.recipients,
            "timestamp": self.timestamp,
            "start": iso8601_string_from_datetime(self.start),
            "stale": iso8601_string_from_datetime(self.stale),
            "xml": self.xml,
            "eud": self.eud.to_json() if self.eud else None,
            "alert": self.alert.to_json() if self.alert else None,
            "point": self.point.to_json() if self.point else None,
            "casevac": self.casevac.to_json() if self.casevac else None,
            "video": self.video.to_json() if self.video else None,
            "geochat": self.geochat.to_json() if self.geochat else None,
        }



"""Added uid column to cot table

Revision ID: 00546817d518
Revises: 807c3ca8e7d0
Create Date: 2024-10-11 20:12:20.866452

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "00546817d518"
down_revision = "807c3ca8e7d0"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("cot", schema=None) as batch_op:
        batch_op.add_column(sa.Column("uid", sa.String(length=255), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("cot", schema=None) as batch_op:
        batch_op.drop_column("uid")

    # ### end Alembic commands ###



"""Added eud_stats table

Revision ID: 1041fae84708
Revises: 21fb5a21f356
Create Date: 2024-12-17 04:37:54.374451

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1041fae84708"
down_revision = "21fb5a21f356"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "eud_stats",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("timestamp", sa.DateTime(), nullable=True),
        sa.Column("eud_uid", sa.String(length=255), nullable=False),
        sa.Column("heap_free_size", sa.Integer(), nullable=True),
        sa.Column("app_framerate", sa.Integer(), nullable=True),
        sa.Column("storage_total", sa.Integer(), nullable=True),
        sa.Column("heap_current_size", sa.Integer(), nullable=True),
        sa.Column("battery", sa.Integer(), nullable=True),
        sa.Column("battery_temp", sa.Integer(), nullable=True),
        sa.Column("deviceDataRx", sa.Integer(), nullable=True),
        sa.Column("heap_max_size", sa.Integer(), nullable=True),
        sa.Column("storage_available", sa.Integer(), nullable=True),
        sa.Column("deviceDataTx", sa.Integer(), nullable=True),
        sa.Column("ip_address", sa.String(length=255), nullable=True),
        sa.Column("battery_status", sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(["eud_uid"], ["euds.uid"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade():
    op.drop_table("eud_stats")
    # ### end Alembic commands ###



"""Fixed video_streams table

Revision ID: 298430e7849d
Revises: f107b45529ba
Create Date: 2024-10-28 19:14:27.017319

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "298430e7849d"
down_revision = "f107b45529ba"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table("video_streams", schema=None) as batch_op:
        batch_op.alter_column(
            "xml", existing_type=sa.VARCHAR(length=255), type_=sa.TEXT(), existing_nullable=True
        )
        batch_op.alter_column(
            "mediamtx_settings",
            existing_type=sa.VARCHAR(length=255),
            type_=sa.TEXT(),
            existing_nullable=False,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("video_streams", schema=None) as batch_op:
        batch_op.alter_column(
            "mediamtx_settings",
            existing_type=sa.TEXT(),
            type_=sa.VARCHAR(length=255),
            existing_nullable=False,
        )
        batch_op.alter_column(
            "xml", existing_type=sa.TEXT(), type_=sa.VARCHAR(length=255), existing_nullable=True
        )

    # ### end Alembic commands ###



import datetime
from dataclasses import dataclass

from sqlalchemy import DateTime, Float, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db
from opentakserver.functions import iso8601_string_from_datetime


@dataclass
class MissionUID(db.Model):
    __tablename__ = "mission_uids"

    uid: Mapped[str] = mapped_column(String(255), primary_key=True)  # Equals the original CoT's UID
    mission_name: Mapped[str] = mapped_column(
        String(255), ForeignKey("missions.name"), nullable=True
    )
    timestamp: Mapped[datetime] = mapped_column(DateTime, nullable=True)
    creator_uid: Mapped[str] = mapped_column(String(255), nullable=True)
    cot_type: Mapped[str] = mapped_column(String(255), nullable=True)
    callsign: Mapped[str] = mapped_column(String(255), nullable=True)
    iconset_path: Mapped[str] = mapped_column(String(255), nullable=True)
    color: Mapped[str] = mapped_column(String(255), nullable=True)
    latitude: Mapped[float] = mapped_column(Float, nullable=True)
    longitude: Mapped[float] = mapped_column(Float, nullable=True)
    mission = relationship("Mission", back_populates="uids")
    mission_change = relationship("MissionChange", back_populates="uid", uselist=False)

    def serialize(self):
        return {
            "timestamp": self.timestamp,
            "creator_uid": self.creator_uid,
            "cot_type": self.cot_type,
            "callsign": self.callsign,
            "iconset_path": self.iconset_path,
            "color": self.color,
            "latitude": self.latitude,
            "longitude": self.longitude,
        }

    def to_json(self):
        return {
            "data": self.uid,
            "timestamp": iso8601_string_from_datetime(self.timestamp),
            "creatorUid": self.creator_uid if self.creator_uid else "",
            "details": {
                "type": self.cot_type,
                "callsign": self.callsign,
                "iconsetPath": self.iconset_path,
                "color": self.color,
                "location": {"lat": self.latitude, "lon": self.longitude},
            },
        }

    def to_details_json(self):
        return {
            "type": self.cot_type,
            "callsign": self.callsign,
            "iconsetPath": self.iconset_path,
            "color": self.color,
            "location": {"lat": self.latitude, "lon": self.longitude},
        }



import json
import traceback
from urllib.parse import urlparse
from xml.etree.ElementTree import Element, fromstring, tostring

import sqlalchemy
from bs4 import BeautifulSoup
from flask import Blueprint
from flask import current_app as app
from flask import jsonify, request
from flask_babel import gettext
from OpenSSL import crypto
from werkzeug.datastructures import ImmutableMultiDict

from opentakserver import __version__ as version
from opentakserver.blueprints.marti_api.marti_api import verify_client_cert
from opentakserver.extensions import db, logger
from opentakserver.forms.MediaMTXPathConfig import MediaMTXPathConfig
from opentakserver.functions import iso8601_string_from_datetime
from opentakserver.models.EUD import EUD
from opentakserver.models.user import User
from opentakserver.models.VideoStream import VideoStream

video_marti_api = Blueprint("video_marti_api", __name__)


@video_marti_api.route("/Marti/vcm", methods=["GET", "POST"])
def video():
    if request.method == "POST":
        soup = BeautifulSoup(request.data, "xml")
        video_connections = soup.find("videoConnections")

        path = video_connections.find("path").text
        if path.startswith("/"):
            path = path[1:]

        if video_connections:
            v = VideoStream()
            v.protocol = video_connections.find("protocol").text
            v.alias = video_connections.find("alias").text
            v.uid = video_connections.find("uid").text
            v.port = video_connections.find("port").text
            v.rover_port = video_connections.find("roverPort").text
            v.ignore_embedded_klv = (
                video_connections.find("ignoreEmbeddedKLV").text.lower() == "true"
            )
            v.preferred_mac_address = video_connections.find("preferredMacAddress").text
            v.preferred_interface_address = video_connections.find("preferredInterfaceAddress").text
            v.path = path
            v.buffer_time = video_connections.find("buffer").text
            v.network_timeout = video_connections.find("timeout").text
            v.rtsp_reliable = video_connections.find("rtspReliable").text
            path_config = MediaMTXPathConfig(None).serialize()
            path_config["sourceOnDemand"] = False
            v.mediamtx_settings = json.dumps(path_config)

            # Discard username and password for security
            feed = soup.find("feed")
            address = feed.find("address").text
            feed.find("address").string.replace_with(address.split("@")[-1])

            v.xml = str(feed)

            with app.app_context():
                try:
                    db.session.add(v)
                    db.session.commit()
                    logger.debug("Inserted Video")
                except sqlalchemy.exc.IntegrityError as e:
                    db.session.rollback()
                    v = db.session.execute(
                        db.select(VideoStream).filter_by(path=v.path)
                    ).scalar_one()
                    v.protocol = video_connections.find("protocol").text
                    v.alias = video_connections.find("alias").text
                    v.uid = video_connections.find("uid").text
                    v.port = video_connections.find("port").text
                    v.rover_port = video_connections.find("roverPort").text
                    v.ignore_embedded_klv = (
                        video_connections.find("ignoreEmbeddedKLV").text.lower() == "true"
                    )
                    v.preferred_mac_address = video_connections.find("preferredMacAddress").text
                    v.preferred_interface_address = video_connections.find(
                        "preferredInterfaceAddress"
                    ).text
                    v.path = video_connections.find("path").text
                    v.buffer_time = video_connections.find("buffer").text
                    v.network_timeout = video_connections.find("timeout").text
                    v.rtsp_reliable = video_connections.find("rtspReliable").text
                    feed = soup.find("feed")
                    address = feed.find("address").text
                    feed.find("address").replace_with(address.split("@")[-1])

                    v.xml = str(feed)

                    db.session.commit()
                    logger.debug("Updated video")

        return "", 200

    elif request.method == "GET":
        try:
            with app.app_context():
                videos = db.session.execute(db.select(VideoStream)).scalars()
                videoconnections = Element("videoConnections")

                for video in videos:
                    # Make sure videos have the correct address based off of the Flask request and not 127.0.0.1
                    # This also forces all streams to bounce through MediaMTX
                    feed = BeautifulSoup(video.xml, "xml")

                    url = urlparse(request.url_root).hostname
                    path = feed.find("path").text
                    if not path.startswith("/"):
                        path = "/" + path

                    if "iTAK" in request.user_agent.string:
                        url = (
                            feed.find("protocol").text
                            + "://"
                            + url
                            + ":"
                            + feed.find("port").text
                            + path
                        )

                    if feed.find("address"):
                        feed.find("address").string.replace_with(url)
                    else:
                        address = feed.new_tag("address")
                        address.string = url
                        feed.feed.append(address)
                    videoconnections.append(fromstring(str(feed)))

            return tostring(videoconnections), 200
        except BaseException as e:
            logger.error(traceback.format_exc())
            return "", 500


@video_marti_api.route("/Marti/api/video")
def get_videos():
    cert = verify_client_cert()
    if not cert:
        # Shouldn't ever get here since nginx already verifies the cert
        return jsonify({"success": False, "error": gettext("Invalid Certificate")}), 400
    username = None
    for a in cert.get_subject().get_components():
        if a[0].decode("UTF-8") == "CN":
            username = a[1].decode("UTF-8")
            break

    user = app.security.datastore.find_user(username=username)
    videos = db.session.execute(db.select(VideoStream)).scalars()

    video_connections = {"videoConnections": []}
    for video in videos:
        video_connections["videoConnections"].append(video.to_marti_json(user))

    return jsonify(video_connections)


@video_marti_api.route("/Marti/api/video", methods=["POST"])
def add_video():
    cert = verify_client_cert()
    username = None
    for a in cert.get_subject().get_components():
        if a[0].decode("UTF-8") == "CN":
            username = a[1].decode("UTF-8")
            break

    videos = request.json

    for video in videos["videoConnections"]:
        feeds = video.get("feeds")

        for feed in feeds:
            data = ImmutableMultiDict({"path": feed["alias"], "source": feed["url"]})
            mediamtx_config = MediaMTXPathConfig(formdata=data, csrf_enabled=False)

            video_stream = VideoStream()
            video_stream.path = feed["alias"]
            video_stream.uid = video["uuid"]
            video_stream.mediamtx_settings = json.dumps(mediamtx_config.serialize())
            video_stream.username = username
            video_stream.rover_port = -1
            video_stream.ignore_embedded_klv = False
            video_stream.buffer_time = None
            video_stream.rtsp_reliable = 1
            video_stream.network_timeout = 10000
            video_stream.generate_xml(urlparse(request.url_root).hostname)
            db.session.add(video_stream)
            db.session.commit()

    return "", 200


@video_marti_api.route("/Marti/api/video/<uid>")
def get_video(uid):
    cert = verify_client_cert()
    username = None
    for a in cert.get_subject().get_components():
        if a[0].decode("UTF-8") == "CN":
            username = a[1].decode("UTF-8")
            break

    video = db.session.query(VideoStream).filter_by(uid=uid).scalar()
    if not video:
        return jsonify({"success": False, "error": gettext("Video not found")}), 404
    user = db.session.execute(db.session.query(User).filter_by(username=username)).first()
    if not user:
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext("User %(username)s not found", username=username),
                }
            ),
            401,
        )
    user = user[0]
    return video.to_marti_json(user), 200


@video_marti_api.route("/Marti/api/video/<uid>", methods=["DELETE"])
def delete_video(uid):
    video_stream = db.session.execute(db.session.query(VideoStream).filter_by(uid=uid))
    if not video_stream:
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext("Video stream with uid %(uid)s not found", uid=uid),
                }
            ),
            404,
        )

    video_stream = video_stream[0]

    db.session.delete(video_stream)
    db.session.commit()

    return "", 200



"""Added publish_time to packages table

Revision ID: 34d5df698b81
Revises: 34dc96ee805b
Create Date: 2024-09-12 17:47:08.855213

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "34d5df698b81"
down_revision = "34dc96ee805b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("device_profiles", schema=None) as batch_op:
        batch_op.alter_column("preference_key", existing_type=sa.String(length=255), nullable=False)

    with op.batch_alter_table("packages", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "publish_time",
                sa.DateTime(),
                server_default=sa.text("(CURRENT_TIMESTAMP)"),
                nullable=False,
            )
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("packages", schema=None) as batch_op:
        batch_op.drop_column("publish_time")

    with op.batch_alter_table("device_profiles", schema=None) as batch_op:
        batch_op.alter_column("preference_key", existing_type=sa.String(length=255), nullable=True)

    # ### end Alembic commands ###



"""Fixed mission_changes and mission_uids relationship

Revision ID: 21fb5a21f356
Revises: 795ebb9262d8
Create Date: 2024-11-08 14:09:41.129561

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "21fb5a21f356"
down_revision = "795ebb9262d8"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("mission_changes", schema=None) as batch_op:
        batch_op.add_column(sa.Column("mission_uid", sa.String(length=255), nullable=True))
        batch_op.create_foreign_key(
            "mission_change_uid", "mission_uids", ["mission_uid"], ["uid"], ondelete="CASCADE"
        )

    with op.batch_alter_table("mission_uids", schema=None) as batch_op:
        batch_op.drop_constraint("mission_change", type_="foreignkey")
        batch_op.drop_column("mission_change_id")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("mission_uids", schema=None) as batch_op:
        batch_op.add_column(sa.Column("mission_change_id", sa.INTEGER(), nullable=True))
        batch_op.create_foreign_key(
            "mission_change", "mission_changes", ["mission_change_id"], ["id"]
        )

    with op.batch_alter_table("mission_changes", schema=None) as batch_op:
        batch_op.drop_constraint("mission_change_uid", type_="foreignkey")
        batch_op.drop_column("mission_uid")

    # ### end Alembic commands ###



"""Added data sync tables

Revision ID: 5d06227dea50
Revises: 6af2256c568d
Create Date: 2024-10-09 14:35:53.352679

"""

import uuid

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "5d06227dea50"
down_revision = "6af2256c568d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "mission_content",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("keywords", sa.JSON(), nullable=True),
        sa.Column("mime_type", sa.String(length=255), nullable=True),
        sa.Column("name", sa.String(length=255), nullable=True),
        sa.Column("submission_time", sa.DateTime(), nullable=True),
        sa.Column("submitter", sa.String(length=255), nullable=True),
        sa.Column("uid", sa.String(length=255), nullable=False),
        sa.Column("creator_uid", sa.String(length=255), nullable=True),
        sa.Column("hash", sa.String(length=255), nullable=True),
        sa.Column("size", sa.Integer(), nullable=True),
        sa.Column("expiration", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("uid"),
    )
    op.create_table(
        "missions",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("description", sa.String(length=255), nullable=True),
        sa.Column("chat_room", sa.String(length=255), nullable=True),
        sa.Column("base_layer", sa.String(length=255), nullable=True),
        sa.Column("bbox", sa.String(length=255), nullable=True),
        sa.Column("path", sa.String(length=255), nullable=True),
        sa.Column("classification", sa.String(length=255), nullable=True),
        sa.Column("tool", sa.String(length=255), nullable=True),
        sa.Column("group_name", sa.String(length=255), nullable=True),
        sa.Column("default_role", sa.String(length=255), nullable=True),
        sa.Column("keywords", sa.JSON(), nullable=True),
        sa.Column("creator_uid", sa.String(length=255), nullable=True),
        sa.Column("create_time", sa.DateTime(), nullable=True),
        sa.Column("external_data", sa.JSON(), nullable=True),
        sa.Column("feeds", sa.JSON(), nullable=True),
        sa.Column("map_layers", sa.JSON(), nullable=True),
        sa.Column("invite_only", sa.Boolean(), nullable=True),
        sa.Column("expiration", sa.Integer(), nullable=False),
        sa.Column("guid", sa.String(length=255), nullable=True),
        sa.Column("password_protected", sa.Boolean(), nullable=True),
        sa.Column("password", sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint("name"),
    )
    op.create_table(
        "mission_changes",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("isFederatedChange", sa.Boolean(), nullable=False),
        sa.Column("change_type", sa.String(length=255), nullable=False),
        sa.Column("mission_name", sa.String(length=255), nullable=False),
        sa.Column("timestamp", sa.DateTime(), nullable=False),
        sa.Column("creator_uid", sa.String(length=255), nullable=False),
        sa.Column("server_time", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(["mission_name"], ["missions.name"], name="mission_changes"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "mission_content_mission",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("mission_content_id", sa.Integer(), nullable=False),
        sa.Column("mission_name", sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(
            ["mission_content_id"], ["mission_content.id"], name="mission_content"
        ),
        sa.ForeignKeyConstraint(["mission_name"], ["missions.name"], name="mission"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "mission_invitations",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("mission_name", sa.String(length=255), nullable=False),
        sa.Column("client_uid", sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(["mission_name"], ["missions.name"], name="mission_invitations"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "mission_roles",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("clientUid", sa.String(length=255), nullable=False),
        sa.Column("username", sa.String(length=255), nullable=False),
        sa.Column("createTime", sa.DateTime(), nullable=False),
        sa.Column("role_type", sa.String(length=255), nullable=False),
        sa.Column("mission_name", sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(["mission_name"], ["missions.name"], name="mission_role_names"),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("mission_roles")
    op.drop_table("mission_invitations")
    op.drop_table("mission_content_mission")
    op.drop_table("mission_changes")
    op.drop_table("missions")
    op.drop_table("mission_content")
    # ### end Alembic commands ###



"""Added type field to MissionInvitation

Revision ID: f2adbb61ff2f
Revises: 35a075531973
Create Date: 2026-01-23 11:40:37.784648

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f2adbb61ff2f"
down_revision = "35a075531973"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("mission_invitations", schema=None) as batch_op:
        batch_op.add_column(sa.Column("type", sa.String(length=255), nullable=True))
        batch_op.drop_constraint(batch_op.f("euds"), type_="foreignkey")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("mission_invitations", schema=None) as batch_op:
        batch_op.create_foreign_key(batch_op.f("euds"), "euds", ["client_uid"], ["uid"])
        batch_op.drop_column("type")

    # ### end Alembic commands ###



import functools

from flask import Blueprint, request
from flask_security import current_user
from flask_socketio import disconnect

from opentakserver.extensions import logger, socketio

ots_socketio_blueprint = Blueprint("ots_socketio_blueprint", __name__)


def authenticated_only(f):
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        if not current_user.is_authenticated:
            logger.debug("Disconnecting {} from {}".format(request.sid, request.namespace))
            disconnect(request.sid, namespace=request.namespace)
        else:
            return f(*args, **kwargs)

    return wrapped


def administrator_only(f):
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        if not current_user.is_authenticated and not current_user.has_role("administrator"):
            logger.debug("Disconnecting {} from {}".format(request.sid, request.namespace))
            disconnect(request.sid, namespace=request.namespace)
        else:
            return f(*args, **kwargs)

    return wrapped


@socketio.on("connect", namespace="/socket.io")
@authenticated_only
def connect(data):
    logger.debug("got a socketio connection from {}".format(current_user.username))


@socketio.on("message", namespace="/socket.io")
@authenticated_only
def data(data):
    logger.debug("Got a message".format(data))



import bleach
from flask import Blueprint
from flask import current_app as app
from flask import jsonify, request

citrap_api_blueprint = Blueprint("citrap_api_blueprint", __name__)


@citrap_api_blueprint.route("/Marti/api/missions/citrap/subscription", methods=["PUT"])
def citrap_subscription():
    uid = bleach.clean(request.args.get("uid"))
    response = {"version": 3, "type": "com.bbn.marti.sync.model.MissionSubscription", "data": {}}
    return "", 201


@citrap_api_blueprint.route("/Marti/api/citrap")
def citrap():
    return jsonify([])



from datetime import datetime, timezone

from sqlalchemy import Boolean, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db
from opentakserver.forms.device_profile_form import DeviceProfileForm
from opentakserver.functions import iso8601_string_from_datetime

# from .EUD import EUD


class DeviceProfiles(db.Model):
    __tablename__ = "device_profiles"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    preference_key: Mapped[str] = mapped_column(String(255), nullable=False)
    preference_value: Mapped[str] = mapped_column(String(255), nullable=True)
    value_class: Mapped[str] = mapped_column(String(255), nullable=True)
    enrollment: Mapped[bool] = mapped_column(Boolean, default=True, nullable=True)
    connection: Mapped[bool] = mapped_column(Boolean, default=False, nullable=True)
    tool: Mapped[str] = mapped_column(String(255), nullable=True)
    active: Mapped[bool] = mapped_column(Boolean, default=True)
    publish_time: Mapped[datetime] = mapped_column(DateTime)
    eud_uid: Mapped[str] = mapped_column(
        String(255),
        ForeignKey("euds.uid", ondelete="CASCADE"),
        nullable=True,
    )
    eud = relationship("EUD", back_populates="profiles")

    def from_wtf(self, form: DeviceProfileForm):
        self.preference_key = form.preference_key.data
        self.preference_value = form.preference_value.data
        self.value_class = f"class java.lang.{form.value_class.data.split('.')[-1]}"
        self.enrollment = form.enrollment.data
        self.connection = form.connection.data
        self.tool = form.tool.data
        self.active = form.active.data
        self.publish_time = datetime.now(timezone.utc)
        self.eud_uid = form.eud_uid.data

    def serialize(self):
        return {
            "preference_key": self.preference_key,
            "preference_value": self.preference_value,
            "value_class": self.value_class,
            "enrollment": self.enrollment,
            "connection": self.connection,
            "tool": self.tool,
            "active": self.active,
            "publish_time": self.publish_time,
            "eud_uid": self.eud_uid,
        }

    def to_json(self):
        return_value = self.serialize()
        return_value["publish_time"] = iso8601_string_from_datetime(self.publish_time)
        return_value["callsign"] = self.eud.callsign if self.eud else None
        return return_value



from dataclasses import dataclass

from flask_security.models import fsqla_v3 as fsqla
from sqlalchemy import String
from sqlalchemy.orm import relationship

from opentakserver.extensions import db
from opentakserver.models.Group import Group

# Leave this import here
from opentakserver.models.Token import Token
from opentakserver.models.WebAuthn import WebAuthn


@dataclass
class User(db.Model, fsqla.FsUserMixin):
    email = db.Column(String(255), nullable=True)
    video_streams = relationship("VideoStream", back_populates="user")
    euds = relationship("EUD", back_populates="user")
    data_packages = relationship("DataPackage", back_populates="user")
    certificate = relationship("Certificate", back_populates="user")
    mission_invitations = relationship("MissionInvitation", back_populates="user")
    tokens = relationship("Token", back_populates="user")
    groups = relationship(
        "Group",
        secondary="groups_users",
        viewonly=True,
        back_populates="users",
        cascade="all, delete",
    )
    group_memberships = relationship("GroupUser", back_populates="user", cascade="all, delete")

    def serialize(self):
        return {
            "id": self.id,
            "username": self.username,
            "active": self.active,
            "last_login_at": self.last_login_at,
            "last_login_ip": self.last_login_ip,
            "current_login_at": self.current_login_at,
            "current_login_ip": self.current_login_ip,
            "email": self.email,
            "login_count": self.login_count,
            "euds": [eud.serialize() for eud in self.euds],
            "video_streams": [v.serialize() for v in self.video_streams],
            "roles": [role.serialize() for role in self.roles],
            "groups": [group.serialize() for group in self.groups],
            "group_memberships": [membership.to_json() for membership in self.group_memberships],
        }

    def to_json(self):
        response = self.serialize()
        response["token"] = self.get_auth_token()
        return response



"""Added device_profile and package tables

Revision ID: 414b87c32cc2
Revises: d91957bb59a0
Create Date: 2024-09-05 19:17:20.979466

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "414b87c32cc2"
down_revision = "d91957bb59a0"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "device_profiles",
        sa.Column("preference_key", sa.String(length=255), nullable=False),
        sa.Column("preference_value", sa.String(length=255), nullable=False),
        sa.Column("value_class", sa.String(length=255), nullable=False),
        sa.Column("enrollment", sa.Boolean(), nullable=True),
        sa.Column("connection", sa.Boolean(), nullable=True),
        sa.Column("tool", sa.String(length=255), nullable=True),
        sa.Column("active", sa.Boolean(), nullable=False),
        sa.Column(
            "publish_time",
            sa.DateTime(),
            server_default=sa.text("(CURRENT_TIMESTAMP)"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("preference_key"),
    )
    op.create_table(
        "packages",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("platform", sa.String(length=255), nullable=False),
        sa.Column("plugin_type", sa.String(length=255), nullable=False),
        sa.Column("package_name", sa.String(length=255), nullable=False),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column("file_name", sa.String(length=255), nullable=False),
        sa.Column("version", sa.String(length=255), nullable=False),
        sa.Column("revision_code", sa.Integer(), nullable=True),
        sa.Column("description", sa.Integer(), nullable=True),
        sa.Column("apk_hash", sa.Integer(), nullable=True),
        sa.Column("os_requirement", sa.Integer(), nullable=True),
        sa.Column("tak_prereq", sa.Integer(), nullable=True),
        sa.Column("file_size", sa.Integer(), nullable=False),
        sa.Column("icon", sa.LargeBinary(), nullable=True),
        sa.Column("icon_filename", sa.String(length=255), nullable=True),
        sa.Column("install_on_enrollment", sa.Boolean(), nullable=True),
        sa.Column("install_on_connection", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("package_name"),
    )
    with op.batch_alter_table("data_packages", schema=None) as batch_op:
        batch_op.add_column(sa.Column("install_on_enrollment", sa.Boolean(), nullable=True))
        batch_op.add_column(sa.Column("install_on_connection", sa.Boolean(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("data_packages", schema=None) as batch_op:
        batch_op.drop_column("install_on_connection")
        batch_op.drop_column("install_on_enrollment")

    op.drop_table("packages")
    op.drop_table("device_profiles")
    # ### end Alembic commands ###



from apscheduler.triggers.combining import AndTrigger, OrTrigger
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers.interval import IntervalTrigger
from flask import Blueprint, jsonify, request
from flask_apscheduler import api
from flask_security import roles_accepted

from opentakserver.extensions import apscheduler

scheduler_api_blueprint = Blueprint("schedule_api_blueprint", __name__)


@scheduler_api_blueprint.route("/api/scheduler/", strict_slashes=False)
@roles_accepted("administrator")
def scheduler_info():
    return api.get_scheduler_info()


@scheduler_api_blueprint.route("/api/scheduler/jobs", strict_slashes=False)
@roles_accepted("administrator")
def get_jobs():
    return api.get_jobs()


@scheduler_api_blueprint.route("/api/scheduler/job/pause", methods=["POST"], strict_slashes=False)
@roles_accepted("administrator")
def pause_job():
    if "job_id" not in request.json:
        return (
            {"success": False, "errors": "Please provide a job_id"},
            400,
            {"Content-Type": "application/json"},
        )
    try:
        return api.pause_job(request.json["job_id"])
    except BaseException as e:
        return {"success": False, "error": str(e)}, 400, {"Content-Type": "application/json"}


@scheduler_api_blueprint.route("/api/scheduler/job/resume", methods=["POST"], strict_slashes=False)
@roles_accepted("administrator")
def resume_job():
    if "job_id" not in request.json:
        return (
            {"success": False, "errors": "Please provide a job_id"},
            400,
            {"Content-Type": "application/json"},
        )
    try:
        #  0 == STATE_STOPPED
        if apscheduler.state == 0:
            apscheduler.start()
        #  2 == STATE_PAUSED
        elif apscheduler.state == 2:
            apscheduler.resume()

        return api.resume_job(request.json["job_id"])
    except BaseException as e:
        return {"success": False, "error": str(e)}, 400, {"Content-Type": "application/json"}


@scheduler_api_blueprint.route("/api/scheduler/job/run", methods=["POST"], strict_slashes=False)
@roles_accepted("administrator")
def run_job():
    if "job_id" not in request.json:
        return (
            {"success": False, "errors": "Please provide a job_id"},
            400,
            {"Content-Type": "application/json"},
        )
    try:
        return api.run_job(request.json["job_id"])
    except BaseException as e:
        return {"success": False, "error": str(e)}, 400, {"Content-Type": "application/json"}


@scheduler_api_blueprint.route("/api/scheduler/job/modify", methods=["POST"], strict_slashes=False)
@roles_accepted("administrator")
def modify_job():
    if "job_id" not in request.json:
        return (
            {"success": False, "errors": "Please provide a job_id"},
            400,
            {"Content-Type": "application/json"},
        )
    try:
        job_id = request.json.pop("job_id")
        trigger = request.json.pop("trigger")
        job = apscheduler.get_job(job_id)
        if trigger == "interval":
            trigger = IntervalTrigger(**request.json)
        elif trigger == "cron":
            trigger = CronTrigger(**request.json)
        elif trigger == "date":
            trigger = DateTrigger(**request.json)
        elif trigger == "and":
            trigger = AndTrigger(**request.json)
        elif trigger == "or":
            trigger = OrTrigger(**request.json)
        else:
            return jsonify({"success": False, "error": "Invalid trigger: {}".format(trigger)}), 400

        job.modify(trigger=trigger)
        return jsonify({"success": True})
    except BaseException as e:
        return {"success": False, "error": str(e)}, 400, {"Content-Type": "application/json"}



"""Added mission_logs table

Revision ID: 14e268184cc5
Revises: d4cc3d4afdb9
Create Date: 2024-10-15 20:28:12.426250

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "14e268184cc5"
down_revision = "d4cc3d4afdb9"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "mission_logs",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("content", sa.String(length=255), nullable=False),
        sa.Column("creator_uid", sa.String(length=255), nullable=False),
        sa.Column("entry_uid", sa.String(length=255), nullable=False),
        sa.Column("mission_name", sa.String(length=255), nullable=False),
        sa.Column("server_time", sa.DateTime(), nullable=False),
        sa.Column("dtg", sa.DateTime(), nullable=False),
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("content_hash", sa.String(length=255), nullable=True),
        sa.Column("keywords", sa.JSON(), nullable=False),
        sa.ForeignKeyConstraint(["mission_name"], ["missions.name"], name="mission_logs"),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("mission_invitations", schema=None) as batch_op:
        batch_op.create_foreign_key("euds", "euds", ["client_uid"], ["uid"])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("mission_invitations", schema=None) as batch_op:
        batch_op.drop_constraint("euds", type_="foreignkey")

    op.drop_table("mission_logs")
    # ### end Alembic commands ###



"""Added eud_uid to device_profiles table, mysql fixes

Revision ID: 61c7cfea0c86
Revises: f2adbb61ff2f
Create Date: 2026-01-28 12:16:08.829691

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = "61c7cfea0c86"
down_revision = "f2adbb61ff2f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("device_profiles", schema=None) as batch_op:
        batch_op.add_column(sa.Column("eud_uid", sa.String(length=255), nullable=True))
        batch_op.create_foreign_key(None, "euds", ["eud_uid"], ["uid"], ondelete="CASCADE")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("device_profiles", schema=None) as batch_op:
        batch_op.drop_constraint(None, type_="foreignkey")
        batch_op.drop_column("eud_uid")

    # ### end Alembic commands ###



import json
import traceback
import xml.etree.ElementTree as ET
from datetime import timedelta, timezone
from uuid import UUID, uuid4

import bleach
import pika
from flask import Blueprint
from flask import current_app as app
from flask import jsonify, request
from flask_babel import gettext
from flask_security import auth_required, current_user
from sqlalchemy import insert, update
from sqlalchemy.exc import IntegrityError

from opentakserver.blueprints.ots_api.api import paginate, route_cot, search
from opentakserver.extensions import db, logger, socketio
from opentakserver.functions import *
from opentakserver.models.CoT import CoT
from opentakserver.models.Marker import Marker
from opentakserver.models.Point import Point

marker_api_blueprint = Blueprint("marker_api_blueprint", __name__)


@marker_api_blueprint.route("/api/markers", methods=["GET"])
@auth_required()
def get_markers():
    query = db.session.query(Marker)
    query = search(query, Marker, "uid")
    query = search(query, Marker, "affiliation")
    query = search(query, Marker, "callsign")

    return paginate(query)


@marker_api_blueprint.route("/api/markers", methods=["POST"])
@auth_required()
def add_marker():
    marker = Marker()
    point = Point()

    try:
        if "latitude" not in request.json.keys() or "longitude" not in request.json.keys():
            return (
                jsonify(
                    {"success": False, "error": gettext("Please provide a latitude and longitude")}
                ),
                400,
            )
        elif float(request.json["latitude"]) < -90 or float(request.json["latitude"]) > 90:
            return (
                jsonify(
                    {
                        "success": False,
                        "error": gettext(
                            "Invalid latitude: %(latitude)s", latitude=request.json["latitude"]
                        ),
                    }
                ),
                400,
            )
        elif float(request.json["longitude"]) < -180 or float(request.json["longitude"]) > 180:
            return (
                jsonify(
                    {
                        "success": False,
                        "error": gettext(
                            "Invalid longitude: %(longitude)s", longitude=request.json["longitude"]
                        ),
                    }
                ),
                400,
            )
    except BaseException as e:
        logger.error(f"Failed to parse lat/lon: {e}")
        return (
            jsonify(
                {"success": False, "error": gettext("Failed to parse lat/lon: %(e)s", e=str(e))}
            ),
            400,
        )

    if "uid" not in request.json.keys():
        return jsonify({"success": False, "error": gettext("Please provide a UID")}), 400
    elif "name" not in request.json.keys():
        return jsonify({"success": False, "error": gettext("Please provide a name")}), 400

    try:
        UUID(request.json["uid"], version=4)
        marker.uid = request.json["uid"]
    except ValueError:
        return (
            jsonify(
                {"success": False, "error": gettext("Invalid UID. UIDs need to be in UUID4 format")}
            ),
            400,
        )

    cot_type = request.json["type"] if "type" in request.json.keys() else "a-u-G"

    marker.affiliation = get_affiliation(cot_type)
    marker.battle_dimension = get_battle_dimension(cot_type)
    marker.mil_std_2525c = cot_type_to_2525c(cot_type)

    if not marker.affiliation or not marker.battle_dimension:
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext("Invalid type: %(cot_type)s", cot_type=cot_type),
                }
            ),
            400,
        )

    if "name" in request.json.keys():
        marker.callsign = bleach.clean(request.json["name"])

    try:
        point.uid = str(uuid4())
        point.device_uid = None
        point.location_source = (
            bleach.clean(request.json["location_source"])
            if "location_source" in request.json.keys()
            else ""
        )
        point.latitude = float(request.json["latitude"])
        point.longitude = float(request.json["longitude"])
        point.course = float(request.json["course"]) if "course" in request.json.keys() else 0
        point.azimuth = float(request.json["azimuth"]) if "azimuth" in request.json.keys() else 0
        point.speed = float(request.json["speed"]) if "speed" in request.json.keys() else 0
        point.battery = float(request.json["battery"]) if "battery" in request.json.keys() else None
        point.fov = float(request.json["fov"]) if "fov" in request.json.keys() else None
        point.ce = float(request.json["ce"]) if "ce" in request.json.keys() else 9999999.0
        point.hae = float(request.json["hae"]) if "hae" in request.json.keys() else 9999999.0
        point.le = float(request.json["le"]) if "le" in request.json.keys() else 9999999.0
        point.timestamp = datetime.now(timezone.utc)

        with app.app_context():
            event = ET.Element("event")
            event.set("type", cot_type)
            event.set("version", "2.0")
            event.set("how", "m-g")
            event.set("uid", marker.uid)
            event.set("time", iso8601_string_from_datetime(point.timestamp))
            event.set("start", iso8601_string_from_datetime(point.timestamp))
            event.set("stale", iso8601_string_from_datetime(point.timestamp + timedelta(days=1)))

            cot_point = ET.SubElement(event, "point")
            cot_point.set("ce", str(point.ce))
            cot_point.set("hae", str(point.hae))
            cot_point.set("le", str(point.le))
            cot_point.set("lat", str(point.latitude))
            cot_point.set("lon", str(point.longitude))

            detail = ET.SubElement(event, "detail")
            detail.set("uid", marker.uid)

            contact = ET.SubElement(detail, "contact")
            contact.set("callsign", marker.callsign)

            track = ET.SubElement(detail, "track")
            track.set("course", str(point.course))
            track.set("speed", str(point.speed))

            if point.azimuth or point.fov:
                sensor = ET.SubElement(detail, "sensor")
                sensor.set("azimuth", str(point.azimuth))
                sensor.set("fov", str(point.fov))
                sensor.set("fovBlue", "1.0")
                sensor.set("fovGreen", "1.0")
                sensor.set("fovRed", "1.0")
                sensor.set("range", "100.0")

            rabbit_credentials = pika.PlainCredentials(
                app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD")
            )
            rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
            rabbit_connection = pika.BlockingConnection(
                pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials)
            )
            channel = rabbit_connection.channel()
            channel.basic_publish(
                exchange="cot_parser",
                routing_key="cot_parser",
                body=json.dumps(
                    {"cot": ET.tostring(event).decode("utf-8"), "uid": app.config["OTS_NODE_ID"]}
                ),
                properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")),
            )
            channel.basic_publish(
                exchange="firehose",
                routing_key="",
                body=json.dumps(
                    {"cot": ET.tostring(event).decode("utf-8"), "uid": app.config["OTS_NODE_ID"]}
                ),
                properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")),
            )
            route_cot(ET.tostring(event).decode("utf-8"), current_user)
            channel.close()
            rabbit_connection.close()

            cot = db.session.execute(
                insert(CoT).values(
                    how="m-g",
                    type=cot_type,
                    timestamp=point.timestamp,
                    xml=ET.tostring(event),
                    start=point.timestamp,
                    stale=point.timestamp + timedelta(days=1),
                    sender_callsign=current_user.username,
                )
            )
            db.session.commit()
            marker.cot_id = cot.inserted_primary_key[0]

            p = db.session.execute(
                insert(Point).values(
                    uid=point.uid,
                    device_uid=point.device_uid,
                    ce=point.ce,
                    hae=point.hae,
                    le=point.le,
                    latitude=point.latitude,
                    longitude=point.longitude,
                    timestamp=point.timestamp,
                    location_source=point.location_source,
                    course=point.course,
                    speed=point.speed,
                    battery=point.battery,
                    fov=point.fov,
                    azimuth=point.azimuth,
                    cot_id=cot.inserted_primary_key[0],
                )
            )
            # Commit here to get the point's auto increment primary key
            db.session.commit()

            marker.point_id = p.inserted_primary_key[0]
            try:
                db.session.add(marker)
                db.session.commit()
            except IntegrityError:
                db.session.rollback()
                db.session.execute(
                    update(Marker)
                    .where(Marker.uid == marker.uid)
                    .values(point_id=marker.point_id, cot_id=marker.cot_id, **marker.serialize())
                )
                db.session.commit()

            # Get the marker with its associated point and cot data
            marker = db.session.execute(db.session.query(Marker).filter_by(uid=marker.uid)).first()[
                0
            ]
            socketio.emit("marker", marker.to_json(), namespace="/socket.io")

            return jsonify({"success": True})

    except BaseException as e:
        logger.error(f"Failed to parse data: {e}")
        logger.error(traceback.format_exc())
        return (
            jsonify({"success": False, "error": gettext("Failed to parse data: %(e)s", e=str(e))}),
            400,
        )


@marker_api_blueprint.route("/api/markers", methods=["DELETE"])
@auth_required()
def delete_marker():
    uid = request.args.get("uid")
    if not uid:
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext("Please provide the UID of the marker to delete"),
                }
            ),
            400,
        )

    with app.app_context():
        query = db.session.query(Marker)
        query = search(query, Marker, "uid")
        marker = db.session.execute(query).first()
        if not marker:
            return jsonify({"success": False, "error": gettext("Unknown UID")}), 404

        marker = marker[0]
        now = datetime.now(timezone.utc)
        event = ET.Element(
            "event",
            {
                "how": "h-g-i-g-o",
                "type": "t-x-d-d",
                "version": "2.0",
                "uid": marker.uid,
                "start": iso8601_string_from_datetime(now),
                "time": iso8601_string_from_datetime(now),
                "stale": iso8601_string_from_datetime(now + timedelta(minutes=10)),
            },
        )
        ET.SubElement(
            event, "point", {"ce": "9999999", "le": "9999999", "hae": "0", "lat": "0", "lon": "0"}
        )
        detail = ET.SubElement(event, "detail")
        ET.SubElement(
            detail, "link", {"relation": "p-p", "uid": marker.uid, "type": marker.cot.type}
        )
        ET.SubElement(
            detail,
            "_flow-tags_",
            {"TAK-Server-f1a8159ef7804f7a8a32d8efc4b773d0": iso8601_string_from_datetime(now)},
        )

        rabbit_credentials = pika.PlainCredentials(
            app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD")
        )
        rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
        rabbit_connection = pika.BlockingConnection(
            pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials)
        )
        channel = rabbit_connection.channel()
        channel.basic_publish(
            exchange="cot_parser",
            routing_key="cot_parser",
            body=json.dumps(
                {"cot": ET.tostring(event).decode("utf-8"), "uid": app.config["OTS_NODE_ID"]}
            ),
            properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")),
        )
        channel.basic_publish(
            exchange="firehose",
            routing_key="",
            body=json.dumps(
                {"cot": ET.tostring(event).decode("utf-8"), "uid": app.config["OTS_NODE_ID"]}
            ),
            properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")),
        )
        route_cot(ET.tostring(event).decode("utf-8"), current_user)
        channel.close()
        rabbit_connection.close()

        db.session.delete(marker)
        db.session.commit()

        return jsonify({"success": True})



import json
import uuid
from dataclasses import dataclass
from urllib.parse import urlparse
from xml.etree.ElementTree import Element, SubElement, tostring

from flask import current_app as app
from flask import request
from sqlalchemy import TEXT, Boolean, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db


@dataclass
class VideoStream(db.Model):
    __tablename__ = "video_streams"

    path: Mapped[str] = mapped_column(String(255), primary_key=True)
    protocol: Mapped[str] = mapped_column(String(255), default="rtsp")
    port: Mapped[int] = mapped_column(Integer, default=8554)
    network_timeout: Mapped[int] = mapped_column(Integer, default=10000)
    uid: Mapped[str] = mapped_column(String(255), nullable=True, default=str(uuid.uuid4()))
    buffer_time: Mapped[int] = mapped_column(Integer, nullable=True)
    rover_port: Mapped[int] = mapped_column(Integer, nullable=True)
    rtsp_reliable: Mapped[int] = mapped_column(Integer, nullable=True, default=1)
    ignore_embedded_klv: Mapped[bool] = mapped_column(Boolean, nullable=True)
    alias: Mapped[str] = mapped_column(String(255), nullable=True)
    preferred_mac_address: Mapped[str] = mapped_column(String(255), nullable=True)
    preferred_interface_address: Mapped[str] = mapped_column(String(255), nullable=True)
    username: Mapped[str] = mapped_column(String(255), ForeignKey("user.username"), nullable=True)
    xml: Mapped[str] = mapped_column(TEXT, nullable=True)
    ready: Mapped[bool] = mapped_column(Boolean, default=False)
    mediamtx_settings: Mapped[str] = mapped_column(TEXT, default="")
    cot_id: Mapped[int] = mapped_column(Integer, ForeignKey("cot.id"), nullable=True)
    cot = relationship("CoT", back_populates="video")
    user = relationship("User", back_populates="video_streams")
    recordings = relationship("VideoRecording", back_populates="video_stream")

    def serialize(self):
        return {
            "network_timeout": self.network_timeout,
            "uid": self.uid,
            "protocol": self.protocol,
            "path": self.path,
            "buffer_time": self.buffer_time,
            "port": self.port,
            "rover_port": self.rover_port,
            "rtsp_reliable": self.rtsp_reliable,
            "ignore_embedded_klv": self.ignore_embedded_klv,
            "alias": self.alias,
            "preferred_mac_address": self.preferred_mac_address,
            "preferred_interface_address": self.preferred_interface_address,
            "username": self.username,
            "ready": self.ready,
        }

    def to_json(self):
        try:
            mediamtx_settings = json.loads(self.mediamtx_settings)
            source = mediamtx_settings["source"]
            record = mediamtx_settings["record"]
        except json.decoder.JSONDecodeError:
            source = None
            record = False

        with app.app_context():
            url = urlparse(request.url_root)
            protocol = url.scheme
            hostname = url.hostname
            port = url.port
            if not port and protocol == "https":
                port = 443
            elif not port and protocol == "http":
                port = 80

            return {
                "network_timeout": self.network_timeout,
                "uid": self.uid,
                "protocol": self.protocol,
                "path": self.path,
                "buffer_time": self.buffer_time,
                "port": self.port,
                "rover_port": self.rover_port,
                "rtsp_reliable": self.rtsp_reliable,
                "ignore_embedded_klv": self.ignore_embedded_klv,
                "alias": self.alias,
                "preferred_mac_address": self.preferred_mac_address,
                "preferred_interface_address": self.preferred_interface_address,
                "username": self.username,
                "ready": self.ready,
                "source": source,
                "record": record,
                "rtsp_link": "rtsp://{}:{}/{}".format(hostname, self.port, self.path),
                "webrtc_link": "{}://{}:{}/webrtc/{}/".format(protocol, hostname, port, self.path),
                "hls_link": "{}://{}:{}/hls/{}/".format(protocol, hostname, port, self.path),
                "thumbnail": f"{protocol}://{hostname}:{port}/api/videos/thumbnail?path={self.path}",
            }

    def to_marti_json(self, user):
        url = urlparse(request.url_root)
        protocol = url.scheme
        hostname = url.hostname
        port = url.port
        if not port and protocol == "https":
            port = 443
        elif not port and protocol == "http":
            port = 80

        video_uuid = str(uuid.uuid4())

        return {
            "active": True,
            "alias": self.path,
            "thumbnail": "",
            "classification": "",
            "feeds": [
                {
                    "uuid": video_uuid,
                    "active": True,
                    "alias": self.path,
                    "url": f"{protocol}://{hostname}:{port}/hls/{self.path}/index.m3u8?jwt={user.get_auth_token()}",
                    "order": 0,
                    "macAddress": self.preferred_mac_address,
                    "roverPort": str(self.rover_port),
                    "ignoreEmbeddedKLV": str(self.ignore_embedded_klv),
                    "source": f"{protocol}://{hostname}:{port}/hls/{self.path}/index.m3u8",
                    "networkTimeout": str(self.network_timeout),
                    "bufferTime": str(self.buffer_time),
                    "rtspReliable": str(self.rtsp_reliable),
                    "thumbnail": "",
                    "classification": "",
                    "latitude": "",
                    "longitude": "",
                    "fov": "",
                    "heading": "",
                    "range": "",
                    "width": 0,
                    "height": 0,
                    "bitrate": 0,
                }
            ],
            "uuid": video_uuid,
        }

    def generate_xml(self, hostname):

        feed = Element("feed")
        # Force rtsp to ensure compatibility with ATAK
        SubElement(feed, "protocol").text = "rtsp"
        SubElement(feed, "alias").text = self.alias if self.alias else self.path
        SubElement(feed, "uid").text = str(self.uid) if self.uid else str(uuid.uuid4())
        SubElement(feed, "address").text = hostname
        SubElement(feed, "port").text = str(self.port) if self.port else "8554"
        SubElement(feed, "roverPort").text = str(self.rover_port)
        SubElement(feed, "ignoreEmbeddedKLV").text = self.ignore_embedded_klv
        SubElement(feed, "preferredMacAddress").text = self.preferred_mac_address
        SubElement(feed, "preferredInterfaceAddress").text = self.preferred_interface_address
        SubElement(feed, "path").text = self.path if self.path else self.alias
        SubElement(feed, "buffer").text = str(self.buffer_time) if self.buffer_time else ""
        SubElement(feed, "timeout").text = (
            str(self.network_timeout) if self.network_timeout else "10000"
        )
        SubElement(feed, "rtspReliable").text = (
            str(self.rtsp_reliable) if self.rtsp_reliable else "1"
        )

        self.xml = tostring(feed).decode("utf-8")



import uuid
from datetime import datetime

from sqlalchemy import DateTime, Float, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db
from opentakserver.forms.point_form import PointForm
from opentakserver.functions import iso8601_string_from_datetime


class Point(db.Model):
    __tablename__ = "points"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    uid: Mapped[str] = mapped_column(String(255))
    device_uid: Mapped[str] = mapped_column(
        String(255), ForeignKey("euds.uid", ondelete="CASCADE"), nullable=True
    )
    latitude: Mapped[float] = mapped_column(Float, nullable=True)
    longitude: Mapped[float] = mapped_column(Float, nullable=True)
    ce: Mapped[float] = mapped_column(Float, nullable=True)
    hae: Mapped[float] = mapped_column(Float, nullable=True)
    le: Mapped[float] = mapped_column(Float, nullable=True)
    course: Mapped[float] = mapped_column(Float, nullable=True)
    speed: Mapped[float] = mapped_column(Float, nullable=True)
    location_source: Mapped[str] = mapped_column(String(255), nullable=True)
    battery: Mapped[float] = mapped_column(Float, nullable=True)
    timestamp: Mapped[datetime] = mapped_column(DateTime)
    azimuth: Mapped[float] = mapped_column(Float, nullable=True)
    # Camera field of view from TAK ICU and OpenTAK ICU
    fov: Mapped[float] = mapped_column(Float, nullable=True)
    cot_id: Mapped[int] = mapped_column(
        Integer, ForeignKey("cot.id", ondelete="CASCADE"), nullable=True
    )
    cot = relationship("CoT", back_populates="point")

    # Only populate this field of the CoT type matches ^a- and how matches either ^m-g or ^h-e
    eud = relationship("EUD", back_populates="points")
    casevac = relationship("CasEvac", cascade="all", back_populates="point")
    geochat = relationship("GeoChat", back_populates="point", uselist=False)
    alert = relationship("Alert", cascade="all", back_populates="point")
    marker: Mapped["Marker"] = relationship(cascade="all, delete", back_populates="point")
    rb_line = relationship("RBLine", cascade="all", back_populates="point")

    def from_wtform(self, form: PointForm):
        self.uid = str(uuid.uuid4())
        self.latitude = form.latitude.data
        self.longitude = form.longitude.data
        self.ce = form.ce.data
        self.hae = form.hae.data
        self.le = form.le.data
        self.course = form.course.data
        self.speed = form.speed.data
        self.location_source = form.location_source.data
        self.battery = form.battery.data
        self.timestamp = form.timestamp.data
        self.azimuth = form.azimuth.data
        self.fov = form.fov.data

    def serialize(self):
        return {
            "uid": self.uid,
            "device_uid": self.device_uid,
            "latitude": self.latitude,
            "longitude": self.longitude,
            "ce": self.ce,
            "hae": self.hae,
            "le": self.le,
            "course": self.course,
            "speed": self.speed,
            "azimuth": self.azimuth,
            "fov": self.fov,
            "location_source": self.location_source,
            "battery": self.battery,
            "timestamp": self.timestamp,
        }

    def to_json(self):
        return {
            "uid": self.uid,
            "device_uid": self.device_uid,
            "latitude": self.latitude,
            "longitude": self.longitude,
            "ce": self.ce,
            "hae": self.hae,
            "le": self.le,
            "course": self.course,
            "speed": self.speed,
            "azimuth": self.azimuth,
            "fov": self.fov,
            "location_source": self.location_source,
            "battery": self.battery,
            "timestamp": iso8601_string_from_datetime(self.timestamp),
            "how": self.cot.how if self.cot else None,
            "type": self.cot.type if self.cot else None,
            "callsign": self.eud.callsign if self.eud else None,
        }



import colorlog
from flask_apscheduler import APScheduler
from flask_babel import Babel
from flask_ldap3_login import LDAP3LoginManager
from flask_mailman import Mail
from flask_migrate import Migrate
from flask_socketio import SocketIO
from flask_sqlalchemy import SQLAlchemy

from opentakserver.models.Base import Base

logger = colorlog.getLogger("OpenTAKServer")

mail = Mail()

apscheduler = APScheduler()

db = SQLAlchemy(model_class=Base)

socketio = SocketIO(async_mode="gevent")

migrate = Migrate()

ldap_manager = LDAP3LoginManager()

babel = Babel()



import datetime
import io
import os
import re
import traceback
import uuid
import zipfile
from urllib.parse import urlparse
from xml.etree.ElementTree import Element, SubElement, tostring

from flask import Blueprint, Response
from flask import current_app as app
from flask import request
from sqlalchemy import or_
from werkzeug.wsgi import FileWrapper

import opentakserver
from opentakserver.blueprints.marti_api.marti_api import verify_client_cert
from opentakserver.extensions import db, ldap_manager, logger
from opentakserver.models.DataPackage import DataPackage
from opentakserver.models.DeviceProfiles import DeviceProfiles
from opentakserver.models.Packages import Packages
from opentakserver.models.Plugins import Plugins

device_profile_marti_api_blueprint = Blueprint("device_profile_marti_api_blueprint", __name__)


def get_ldap_attributes(prefs: Element):
    if app.config.get("OTS_ENABLE_LDAP"):
        cert = verify_client_cert()
        if not cert:
            return

        username = cert.get_subject().commonName
        user_info = ldap_manager.get_user_info_for_username(username)

        for field in user_info:
            if field.lower() == app.config.get("OTS_LDAP_COLOR_ATTRIBUTE").lower():
                team = SubElement(
                    prefs, "entry", {"key": "locationTeam", "class": "class java.lang.String"}
                )
                team.text = user_info[field][0]

            elif field.lower() == app.config.get("OTS_LDAP_ROLE_ATTRIBUTE").lower():
                role = SubElement(
                    prefs, "entry", {"key": "atakRoleType", "class": "class java.lang.String"}
                )
                role.text = user_info[field][0]

            elif field.lower() == app.config.get("OTS_LDAP_CALLSIGN_ATTRIBUTE").lower():
                callsign = SubElement(
                    prefs, "entry", {"key": "locationCallsign", "class": "class java.lang.String"}
                )
                callsign.text = user_info[field][0]

            elif field.lower().startswith(
                app.config.get("OTS_LDAP_PREFERENCE_ATTRIBUTE_PREFIX").lower()
            ):
                # Remove prefix, case insensitive
                remove_prefix = re.compile(
                    re.escape(app.config.get("OTS_LDAP_PREFERENCE_ATTRIBUTE_PREFIX")), re.IGNORECASE
                )
                attribute = remove_prefix.sub("", field)
                subelement = SubElement(
                    prefs, "entry", {"key": attribute, "class": "class java.lang.String"}
                )
                subelement.text = user_info[field][0]


def create_profile_zip(enrollment=True, syncSecago=-1, clientUid: str | None = None):
    # preference.pref
    prefs = Element("preferences")
    pref = SubElement(prefs, "preference", {"version": "1", "name": "com.atakmap.app_preferences"})

    get_ldap_attributes(pref)

    plugins = None
    data_packages = None
    device_profiles = None
    zip_buffer = io.BytesIO()
    zipf = zipfile.ZipFile(zip_buffer, "a", zipfile.ZIP_DEFLATED, False)

    # MANIFEST file
    manifest = Element("MissionPackageManifest", {"version": "2"})
    config = SubElement(manifest, "Configuration")
    SubElement(config, "Parameter", {"name": "uid", "value": str(uuid.uuid4())})
    SubElement(config, "Parameter", {"name": "name", "value": "Device Profile"})
    SubElement(config, "Parameter", {"name": "onReceiveDelete", "value": "true"})

    contents = SubElement(manifest, "Contents")
    SubElement(
        contents,
        "Content",
        {"ignore": "false", "zipEntry": "5c2bfcae3d98c9f4d262172df99ebac5/preference.pref"},
    )
    SubElement(
        contents,
        "Content",
        {"ignore": "false", "zipEntry": "5c2bfcae3d98c9f4d262172df99ebac5/truststore-root.p12"},
    )

    enable_update_server = SubElement(
        pref, "entry", {"key": "appMgmtEnableUpdateServer", "class": "class java.lang.Boolean"}
    )
    enable_update_server.text = "true"

    update_server_address = SubElement(
        pref, "entry", {"key": "atakUpdateServerUrl", "class": "class java.lang.String"}
    )
    update_server_address.text = f"https://{urlparse(request.url_root).hostname}:{app.config.get('OTS_MARTI_HTTPS_PORT')}/api/packages"

    startup_sync = SubElement(
        pref, "entry", {"key": "repoStartupSync", "class": "class java.lang.Boolean"}
    )
    startup_sync.text = "true"

    enable_profiles = SubElement(
        pref, "entry", {"key": "deviceProfileEnableOnConnect", "class": "class java.lang.Boolean"}
    )
    enable_profiles.text = "true"

    ca_location = SubElement(
        pref, "entry", {"key": "updateServerCaLocation", "class": "class java.lang.String"}
    )
    ca_location.text = "/storage/emulated/0/atak/cert/truststore-root.p12"

    ca_password = SubElement(
        pref, "entry", {"key": "updateServerCaPassword", "class": "class java.lang.String"}
    )
    ca_password.text = app.config.get("OTS_CA_PASSWORD")

    enable_channels_host = SubElement(
        pref,
        "entry",
        {
            "key": f"prefs_enable_channels_host-{urlparse(request.url_root).hostname}",
            "class": "class java.lang.String",
        },
    )
    enable_channels_host.text = "true"

    enable_channels = SubElement(
        pref, "entry", {"key": "prefs_enable_channels", "class": "class java.lang.String"}
    )
    enable_channels.text = "true" if app.config.get("OTS_ENABLE_CHANNELS") else "false"

    if enrollment:
        # Add the maps to the zip
        if app.config.get("OTS_PROFILE_MAP_SOURCES") and enrollment:
            maps_path = os.path.join(os.path.dirname(opentakserver.__file__), "maps")
            for root, dirs, map_files in os.walk(maps_path):
                for map in map_files:
                    zipf.writestr(f"maps/{map}", open(os.path.join(maps_path, map), "r").read())
                    SubElement(contents, "Content", {"ignore": "false", "zipEntry": f"maps/{map}"})

            device_profiles = db.session.execute(
                db.session.query(DeviceProfiles).filter_by(enrollment=True, active=True)
            ).all()
            plugins = db.session.execute(
                db.session.query(Packages).filter_by(install_on_enrollment=True)
            ).all()
            data_packages = db.session.execute(
                db.session.query(DataPackage).filter_by(install_on_enrollment=True)
            ).all()
    else:
        device_profile_query = db.session.query(DeviceProfiles).filter_by(
            connection=True, active=True
        )
        data_package_query = db.session.query(DataPackage).filter_by(install_on_connection=True)
        plugins_query = db.session.query(Packages).filter_by(install_on_connection=True)

        if syncSecago > 0:
            publish_time = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(
                seconds=syncSecago
            )
            device_profile_query = device_profile_query.filter(
                DeviceProfiles.publish_time >= publish_time
            )
            data_package_query = data_package_query.filter(
                DataPackage.submission_time >= publish_time
            )
            plugins_query = plugins_query.filter(Packages.publish_time >= publish_time)
        if clientUid:
            device_profile_query = device_profile_query.filter(
                or_(DeviceProfiles.eud_uid == clientUid, DeviceProfiles.eud_uid is None)
            )
            # TODO: Support data packages and plugins per EUD
        else:
            device_profile_query = device_profile_query.filter(DeviceProfiles.clientUid is None)

        device_profiles = db.session.execute(device_profile_query).all()
        data_packages = db.session.execute(data_package_query).all()
        plugins = db.session.execute(plugins_query).all()

    for profile in device_profiles:
        p = SubElement(
            pref, "entry", {"key": profile[0].preference_key, "class": profile[0].value_class}
        )
        p.text = profile[0].preference_value

    if plugins:
        for plugin in plugins:
            plugin = plugin[0]

            SubElement(
                contents,
                "Content",
                {
                    "ignore": "false",
                    "zipEntry": f"5c2bfcae3d98c9f4d262172df99ebac5/{plugin.file_name}",
                },
            )
            zipf.write(
                os.path.join(app.config.get("OTS_DATA_FOLDER"), "packages", plugin.file_name),
                f"5c2bfcae3d98c9f4d262172df99ebac5/{plugin.file_name}",
            )

    if data_packages:
        for data_package in data_packages:
            data_package = data_package[0]
            SubElement(
                contents,
                "Content",
                {
                    "ignore": "false",
                    "zipEntry": f"5c2bfcae3d98c9f4d262172df99ebac5/{data_package.filename}",
                },
            )
            zipf.write(
                os.path.join(app.config.get("UPLOAD_FOLDER"), f"{data_package.hash}.zip"),
                f"5c2bfcae3d98c9f4d262172df99ebac5/{data_package.filename}",
            )

    zipf.writestr("MANIFEST/manifest.xml", tostring(manifest))

    zipf.writestr("5c2bfcae3d98c9f4d262172df99ebac5/preference.pref", tostring(prefs))

    with open(
        os.path.join(app.config.get("OTS_CA_FOLDER"), "truststore-root.p12"), "rb"
    ) as truststore:
        zipf.writestr("5c2bfcae3d98c9f4d262172df99ebac5/truststore-root.p12", truststore.read())

    return zip_buffer


# Authentication for /Marti endpoints handled by client cert validation
# EUDs hit this endpoint after a successful certificate enrollment
@device_profile_marti_api_blueprint.route("/Marti/api/tls/profile/enrollment")
def enrollment_profile():
    try:
        profile_zip = create_profile_zip()
        profile_zip.seek(0)
        return Response(
            FileWrapper(profile_zip), mimetype="application/zip", direct_passthrough=True
        )
    except BaseException as e:
        logger.error(f"Failed to send enrollment package: {e}")
        logger.debug(traceback.format_exc())
        return "", 500


# EUDs hit this endpoint when the app connects to the server if repoStartupSync is enabled
@device_profile_marti_api_blueprint.route("/api/connection")
@device_profile_marti_api_blueprint.route("/Marti/api/device/profile/connection")
def connection_profile():
    try:
        syncSecago = -1
        client_uid = None
        if "syncSecago" in request.args:
            try:
                syncSecago = int(request.args["syncSecago"])
            except ValueError:
                pass
        if "clientUid" in request.args:
            client_uid = request.args["clientUid"]

        profile_zip = create_profile_zip(False, syncSecago, client_uid)
        profile_zip.seek(0)
        return Response(
            FileWrapper(profile_zip), mimetype="application/zip", direct_passthrough=True
        )
    except BaseException as e:
        logger.error(f"Failed to send enrollment package: {e}")
        logger.debug(traceback.format_exc())
        return "", 500



import base64
import datetime
import json
import os
import traceback
import uuid
from xml.etree.ElementTree import Element, SubElement, tostring

import pika
import sqlalchemy.exc
import unishox2
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from google.protobuf.json_format import MessageToJson
from meshtastic import BROADCAST_NUM, mesh_pb2, mqtt_pb2, portnums_pb2, protocols
from pika.spec import Basic, BasicProperties, Channel

from opentakserver.controllers.rabbitmq_client import RabbitMQClient
from opentakserver.extensions import db, logger, socketio
from opentakserver.models.EUD import EUD
from opentakserver.models.Meshtastic import MeshtasticChannel
from opentakserver.models.Point import Point
from opentakserver.models.Team import Team
from opentakserver.proto import atak_pb2


class MeshtasticController(RabbitMQClient):
    def __init__(self, context):
        super().__init__(context)
        self.node_names = {}
        self.logger.info("Starting Meshtastic controller...")
        self.meshtastic_devices = {}
        self.get_euds()
        self.get_channels()

    def get_euds(self):
        with self.context:
            euds = self.db.session.execute(self.db.session.query(EUD)).scalars()
            for eud in euds:
                meshtastic_id = eud.meshtastic_id
                if not eud.meshtastic_id:
                    eud.meshtastic_id = int.from_bytes(os.urandom(4), "big")
                    self.db.session.add(eud)
                    self.db.session.commit()
                self.meshtastic_devices[eud.uid] = {
                    "hw_model": eud.device,
                    "long_name": eud.callsign,
                    "short_name": "",
                    "firmware_version": eud.version,
                    "last_lat": "0.0",
                    "last_lon": "0.0",
                    "battery": 0,
                    "meshtastic_id": "{:x}".format(eud.meshtastic_id),
                    "voltage": 0,
                    "uptime": 0,
                    "last_alt": "9999999.0",
                    "course": "0.0",
                    "speed": "0.0",
                    "team": "Cyan",
                    "role": "Team Member",
                    "uid": eud.uid,
                    "macaddr": eud.meshtastic_macaddr,
                }

    def get_channels(self):
        with self.context:
            channels = self.db.session.execute(self.db.session.query(MeshtasticChannel)).scalars()
            downlink_channels = {}
            for channel in channels:
                if channel.downlink_enabled:
                    downlink_channels[channel.name] = channel.psk
            self.context.app.config.update({"OTS_MESHTASTIC_DOWNLINK_CHANNELS": downlink_channels})

    def on_channel_open(self, channel):
        self.rabbit_channel = channel
        self.rabbit_channel.queue_declare(queue="meshtastic")
        self.rabbit_channel.queue_bind(exchange="amq.topic", queue="meshtastic", routing_key="#")
        self.rabbit_channel.basic_consume(
            queue="meshtastic", on_message_callback=self.on_message, auto_ack=True
        )
        self.rabbit_channel.add_on_close_callback(self.on_close)

    def decrypt(self, mp: mesh_pb2.MeshPacket, channel_name: str) -> mesh_pb2.MeshPacket:
        if channel_name not in self.context.app.config.get("OTS_MESHTASTIC_DOWNLINK_CHANNELS"):
            logger.error(
                f"{channel_name} not in {self.context.app.config.get('OTS_MESHTASTIC_DOWNLINK_CHANNELS')}"
            )
            return mp

        if not mp.HasField("encrypted"):
            self.logger.warning("Packet has no 'encrypted' field")
            return mp

        # Make a new MeshPacket with no encrypted field. The decoded field will be set after decryption.
        decrypted_mesh_packet = mesh_pb2.MeshPacket()
        setattr(decrypted_mesh_packet, "from", getattr(mp, "from"))
        decrypted_mesh_packet.to = mp.to
        decrypted_mesh_packet.channel = mp.channel
        decrypted_mesh_packet.id = mp.id
        decrypted_mesh_packet.rx_time = mp.rx_time
        decrypted_mesh_packet.rx_snr = mp.rx_snr
        decrypted_mesh_packet.hop_limit = mp.hop_limit
        decrypted_mesh_packet.want_ack = mp.want_ack
        decrypted_mesh_packet.priority = mp.priority
        decrypted_mesh_packet.rx_rssi = mp.rx_rssi
        decrypted_mesh_packet.delayed = mp.delayed
        decrypted_mesh_packet.via_mqtt = mp.via_mqtt
        decrypted_mesh_packet.hop_start = mp.hop_start
        decrypted_mesh_packet.public_key = mp.public_key
        decrypted_mesh_packet.pki_encrypted = mp.pki_encrypted
        decrypted_mesh_packet.next_hop = mp.next_hop
        decrypted_mesh_packet.relay_node = mp.relay_node
        decrypted_mesh_packet.tx_after = mp.tx_after
        decrypted_mesh_packet.transport_mechanism = mp.transport_mechanism

        psk = self.context.app.config.get("OTS_MESHTASTIC_DOWNLINK_CHANNELS")[channel_name]
        # Use the full default encryption key
        if psk == "AQ==":
            psk = "1PG7OiApB1nwvP+rz05pAQ=="

        # Get the channel key from the DB
        key_bytes = base64.b64decode(psk.encode("ascii"))

        nonce = getattr(mp, "id").to_bytes(8, "little") + getattr(mp, "from").to_bytes(8, "little")
        cipher = Cipher(algorithms.AES(key_bytes), modes.CTR(nonce), backend=default_backend())
        decryptor = cipher.decryptor()
        decrypted_bytes = decryptor.update(getattr(mp, "encrypted")) + decryptor.finalize()

        data = mesh_pb2.Data()
        data.ParseFromString(decrypted_bytes)
        decrypted_mesh_packet.decoded.CopyFrom(data)

        self.logger.debug("Decryption successful")

        return decrypted_mesh_packet

    def on_message(
        self,
        unused_channel: Channel,
        basic_deliver: Basic.Deliver,
        properties: BasicProperties,
        body,
    ):
        # Don't process outgoing message from TAK EUDs to the Meshtastic Network, only messages from the Meshtastic
        # network to TAK EUDs
        if basic_deliver.routing_key.endswith("outgoing"):
            return

        try:
            meshtastic_channel_name = basic_deliver.routing_key.split(".")[3]
        except IndexError:
            return

        se = mqtt_pb2.ServiceEnvelope()

        try:
            se.ParseFromString(body)
            mp = se.packet
        except Exception as e:
            self.logger.error(f"ERROR: parsing service envelope: {str(e)}")
            self.logger.error(f"{body}")
            return

        meshtastic_id = getattr(mp, "from")
        meshtastic_id = f"{meshtastic_id:08x}"
        to_id = mp.to
        if to_id == BROADCAST_NUM:
            to_id = "all"
        else:
            to_id = f"{to_id:08x}"

        pn = portnums_pb2.PortNum.Name(mp.decoded.portnum)

        prefix = f"{mp.channel} [{meshtastic_id}->{to_id}] {pn}:"
        if mp.HasField("encrypted") and not mp.HasField("decoded"):
            try:
                mp = self.decrypt(mp, meshtastic_channel_name)
                pn = portnums_pb2.PortNum.Name(mp.decoded.portnum)
                prefix = f"{mp.channel} [{meshtastic_id}->{to_id}] {pn}:"
            except Exception as e:
                self.logger.warning(f"{prefix} could not be decrypted: {e}")
                self.logger.debug(traceback.format_exc())
                return

        handler = protocols.get(mp.decoded.portnum)
        if handler is None:
            try:
                if portnums_pb2.PortNum.Name(mp.decoded.portnum) == "ATAK_PLUGIN":
                    tak_packet = atak_pb2.TAKPacket()
                    tak_packet.ParseFromString(mp.decoded.payload)
                    self.protobuf_to_cot(tak_packet, meshtastic_id, to_id, pn, meshtastic_id)
                    self.logger.info(tak_packet)
            except:
                self.logger.error(traceback.format_exc())

            return

        if handler.protobufFactory is None:
            self.logger.debug(f"{prefix} {mp}")
            self.protobuf_to_cot(mp.decoded.payload, meshtastic_id, to_id, pn, meshtastic_id)
            self.logger.info(mp.decoded.payload)
        else:
            try:
                pb = handler.protobufFactory()
                pb.ParseFromString(mp.decoded.payload)
                p = MessageToJson(pb)
                if mp.decoded.portnum == portnums_pb2.PortNum.NODEINFO_APP:
                    self.node_names[getattr(mp, "from")] = pb.short_name
                    prefix = f"{mp.channel} [{meshtastic_id}->{to_id}] {pn}:"
                    self.rabbit_channel.queue_declare(queue=meshtastic_id)
                self.logger.debug(f"{prefix} {p}")
                self.protobuf_to_cot(pb, meshtastic_id, to_id, pn, meshtastic_id)
                self.logger.info(pb)
            except:
                self.logger.error(traceback.format_exc())

        # Forward this Meshtastic message to other Meshtastic channels which have downlink enabled
        for channel in self.context.app.config.get("OTS_MESHTASTIC_DOWNLINK_CHANNELS"):
            # Don't re-publish back to the same mesh
            if channel == meshtastic_channel_name:
                continue
            routing_key = "{}.2.e.{}.".format(
                self.context.app.config.get("OTS_MESHTASTIC_TOPIC"), channel
            )
            if not basic_deliver.routing_key.startswith(routing_key):
                self.rabbit_channel.basic_publish(
                    exchange="amq.topic",
                    routing_key=routing_key + "outgoing",
                    body=body,
                    properties=pika.BasicProperties(
                        expiration=self.context.app.config.get("OTS_RABBITMQ_TTL")
                    ),
                )

    def cot(self, pb, from_id, to_id, portnum, how="m-g", cot_type="a-f-G-U-C", uid=None):
        if (
            not uid
            and from_id in self.meshtastic_devices
            and self.meshtastic_devices[from_id]["uid"]
        ):
            uid = self.meshtastic_devices[from_id]["uid"]
        elif not uid:
            uid = from_id

        now = datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
        stale = (
            datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=1)
        ).strftime("%Y-%m-%dT%H:%M:%SZ")

        event = Element(
            "event",
            {
                "how": how,
                "type": cot_type,
                "version": "2.0",
                "uid": uid,
                "start": now,
                "time": now,
                "stale": stale,
            },
        )

        SubElement(
            event,
            "point",
            {
                "ce": "9999999.0",
                "le": "9999999.0",
                "hae": str(self.meshtastic_devices[from_id]["last_alt"]),
                "lat": str(self.meshtastic_devices[from_id]["last_lat"]),
                "lon": str(self.meshtastic_devices[from_id]["last_lon"]),
            },
        )

        detail = SubElement(event, "detail")
        if portnum == "TEXT_MESSAGE_APP" or (portnum == "ATAK_PLUGIN" and pb.HasField("chat")):
            return event, detail
        else:
            SubElement(
                detail,
                "takv",
                {
                    "device": self.meshtastic_devices[from_id]["hw_model"],
                    "version": self.meshtastic_devices[from_id]["firmware_version"],
                    "platform": "Meshtastic",
                    "os": "Meshtastic",
                    "macaddr": self.meshtastic_devices[from_id]["macaddr"],
                    "meshtastic_id": self.meshtastic_devices[from_id]["meshtastic_id"],
                },
            )
            SubElement(
                detail,
                "contact",
                {"callsign": self.meshtastic_devices[from_id]["long_name"], "endpoint": "MQTT"},
            )
            SubElement(detail, "uid", {"Droid": self.meshtastic_devices[from_id]["long_name"]})
            SubElement(detail, "precisionlocation", {"altsrc": "GPS", "geopointsrc": "GPS"})
            SubElement(
                detail, "status", {"battery": str(self.meshtastic_devices[from_id]["battery"])}
            )
            SubElement(detail, "track", {"course": "0.0", "speed": "0.0"})
            SubElement(
                detail,
                "__group",
                {
                    "name": self.meshtastic_devices[from_id]["team"],
                    "role": self.meshtastic_devices[from_id]["role"],
                },
            )

        return event

    def insert_or_update_eud(self, uid: str, from_id: str, update_if_exists: bool = True):
        eud = EUD()
        eud.uid = uid
        eud.callsign = self.meshtastic_devices[from_id]["long_name"]
        eud.device = self.meshtastic_devices[from_id]["hw_model"]
        eud.os = "Meshtastic"
        eud.platform = "Meshtastic"
        eud.version = self.meshtastic_devices[from_id]["firmware_version"]
        eud.team_role = self.meshtastic_devices[from_id]["role"]
        eud.meshtastic_id = int(self.meshtastic_devices[from_id]["meshtastic_id"], 16)
        eud.meshtastic_macaddr = self.meshtastic_devices[from_id]["macaddr"]

        with self.context:
            socketio.emit("eud", eud.to_json(), namespace="/socket.io")

        with self.context:
            team = db.session.execute(
                db.session.query(Team).filter_by(name=self.meshtastic_devices[from_id]["team"])
            ).scalar()
            if team:
                eud.team_id = team.id

            try:
                db.session.add(eud)
                db.session.commit()
            except sqlalchemy.exc.IntegrityError:
                db.session.rollback()
                if update_if_exists:
                    db.session.execute(
                        sqlalchemy.update(EUD).where(EUD.uid == uid).values(**eud.serialize())
                    )
                    db.session.commit()

    def position(self, pb, from_id, to_id, portnum):
        try:
            if (
                portnum == "MAP_REPORT_APP"
                and pb.firmware_version != self.meshtastic_devices[from_id]["firmware_version"]
            ):
                try:
                    with self.context:
                        eud = self.db.session.execute(
                            self.db.session.query(EUD).filter_by(uid=from_id)
                        ).first()[0]
                        eud.version = pb.firmware_version
                        eud.device = pb.hw_model
                        eud.callsign = pb.long_name
                        self.db.session.add(eud)
                        self.db.session.commit()
                        if from_id not in self.meshtastic_devices:
                            self.meshtastic_devices[from_id] = {
                                "hw_model": "",
                                "long_name": f"Meshtastic {str(from_id)[-4:]}",
                                "short_name": str(from_id)[-4:],
                                "macaddr": "",
                                "firmware_version": "",
                                "last_lat": "0.0",
                                "last_lon": "0.0",
                                "battery": 0,
                                "meshtastic_id": "",
                                "voltage": 0,
                                "uptime": 0,
                                "last_alt": "9999999.0",
                                "course": "0.0",
                                "speed": "0.0",
                                "team": "Cyan",
                                "role": "Team Member",
                                "uid": None,
                            }

                        self.meshtastic_devices[from_id]["firmware_version"] = pb.firmware_version
                        self.meshtastic_devices[from_id]["hw_model"] = mesh_pb2.HardwareModel.Name(
                            pb.hw_model
                        )
                        self.meshtastic_devices[from_id]["long_name"] = (
                            pb.long_name or f"Meshtastic {str(from_id)}"
                        )
                        self.meshtastic_devices[from_id]["short_name"] = (
                            pb.short_name or str(from_id)[-4:]
                        )
                except BaseException as e:
                    self.logger.error(
                        "Failed to update {}'s firmware version: {}".format(from_id, e)
                    )

            self.meshtastic_devices[from_id]["last_lat"] = pb.latitude_i * 0.0000001
            self.meshtastic_devices[from_id]["last_lon"] = pb.longitude_i * 0.0000001
            self.meshtastic_devices[from_id]["last_alt"] = pb.altitude
            if portnum == "POSITION_APP":
                self.meshtastic_devices[from_id]["course"] = (
                    pb.ground_track if pb.ground_track else "0.0"
                )
                self.meshtastic_devices[from_id]["speed"] = (
                    pb.ground_speed if pb.ground_speed else "0.0"
                )

            self.insert_or_update_eud(from_id, from_id, False)

            point = Point()
            point.uid = str(uuid.uuid4())
            point.device_uid = from_id
            point.latitude = self.meshtastic_devices[from_id]["last_lat"]
            point.longitude = self.meshtastic_devices[from_id]["last_lon"]
            point.hae = self.meshtastic_devices[from_id]["last_alt"]
            point.course = self.meshtastic_devices[from_id]["course"]
            point.speed = self.meshtastic_devices[from_id]["speed"]
            point.timestamp = datetime.datetime.now(tz=datetime.timezone.utc)

            with self.context:
                db.session.add(point)
                db.session.commit()

                socketio.emit("point", point.to_json(), namespace="/socket.io")

            return self.cot(pb, from_id, to_id, portnum)
        except BaseException as e:
            self.logger.error("Failed to create CoT: {}".format(str(e)))
            self.logger.error(traceback.format_exc())
            return

    def text_message(self, pb, from_id, to_id, portnum):
        callsign = from_id
        if from_id in self.meshtastic_devices:
            callsign = self.meshtastic_devices[from_id]["long_name"]

        chatroom = "All Chat Rooms"
        for meshtastic_device in self.meshtastic_devices:
            meshtastic_device = self.meshtastic_devices[meshtastic_device]
            if meshtastic_device["meshtastic_id"] == to_id:
                chatroom = meshtastic_device["uid"]
                break

        if from_id in self.meshtastic_devices and self.meshtastic_devices[from_id]["uid"]:
            from_uid = self.meshtastic_devices[from_id]["uid"]
        else:
            from_uid = from_id

        message_uid = str(uuid.uuid4())
        event, detail = self.cot(
            pb,
            from_uid,
            chatroom,
            portnum,
            how="h-g-i-g-o",
            cot_type="b-t-f",
            uid="GeoChat.{}.{}.{}".format(from_uid, chatroom, message_uid),
        )

        chat = SubElement(
            detail,
            "__chat",
            {
                "chatroom": chatroom,
                "groupOwner": "false",
                "id": chatroom,
                "messageId": message_uid,
                "parent": "RootContactGroup",
                "senderCallsign": callsign,
            },
        )
        SubElement(chat, "chatgrp", {"id": chatroom, "uid0": from_uid, "uid1": chatroom})
        SubElement(detail, "link", {"relation": "p-p", "type": "a-f-G-U-C", "uid": from_uid})
        remarks = SubElement(
            detail,
            "remarks",
            {
                "source": "BAO.F.ATAK.{}".format(from_uid),
                "time": datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ"),
                "to": chatroom,
            },
        )

        remarks.text = pb.decode("utf-8", "replace")

        return event

    def node_info(self, pb, from_id, to_id, portnum):
        if portnum == "ATAK_PLUGIN":
            uid = unishox2.decompress(pb.contact.device_callsign, len(pb.contact.device_callsign))
            self.meshtastic_devices[from_id]["uid"] = uid
            self.meshtastic_devices[from_id]["long_name"] = unishox2.decompress(
                pb.contact.callsign, len(pb.contact.callsign)
            )
            self.meshtastic_devices[from_id]["short_name"] = uid[-4:]
            self.meshtastic_devices[from_id]["battery"] = pb.status.battery
            if pb.group.team != 0:
                self.meshtastic_devices[from_id]["team"] = atak_pb2.Team.Name(pb.group.team)
            if pb.group.role != 0:
                self.meshtastic_devices[from_id]["role"] = atak_pb2.MemberRole.Name(pb.group.role)

            self.insert_or_update_eud(uid, from_id)
        else:
            hw_model = mesh_pb2.HardwareModel.Name(pb.hw_model)
            self.meshtastic_devices[from_id]["hw_model"] = hw_model if hw_model else ""
            self.meshtastic_devices[from_id]["long_name"] = (
                str(pb.long_name) if pb.long_name else f"Meshtastic {str(from_id)[-4:]}"
            )
            self.meshtastic_devices[from_id]["short_name"] = (
                str(pb.short_name) if pb.short_name else str(from_id)[-4:]
            )
            self.meshtastic_devices[from_id]["macaddr"] = (
                base64.b64encode(pb.macaddr).decode("ascii") if pb.macaddr else ""
            )

            self.insert_or_update_eud(from_id, from_id)

        return self.cot(pb, from_id, to_id, portnum)

    def telemetry(self, pb, from_id, to_id, portnum):
        if pb.HasField("device_metrics"):
            self.meshtastic_devices[from_id]["battery"] = pb.device_metrics.battery_level
            self.meshtastic_devices[from_id]["voltage"] = pb.device_metrics.voltage
            self.meshtastic_devices[from_id]["uptime"] = pb.device_metrics.uptime_seconds
        elif pb.HasField("environment_metrics"):
            self.meshtastic_devices[from_id]["temperature"] = pb.environment_metrics.temperature
            self.meshtastic_devices[from_id][
                "relative_humidity"
            ] = pb.environment_metrics.relative_humidity
            self.meshtastic_devices[from_id][
                "barometric_pressure"
            ] = pb.environment_metrics.barometric_pressure
            self.meshtastic_devices[from_id][
                "gas_resistance"
            ] = pb.environment_metrics.gas_resistance
            self.meshtastic_devices[from_id]["voltage"] = pb.environment_metrics.voltage
            self.meshtastic_devices[from_id]["current"] = pb.environment_metrics.current
            self.meshtastic_devices[from_id]["iaq"] = pb.environment_metrics.iaq

    def atak_plugin(self, pb, from_id, to_id, portnum):
        self.node_info(pb, from_id, to_id, portnum)

        if pb.HasField("status"):
            self.meshtastic_devices[from_id]["battery"] = pb.status.battery

        if pb.HasField("pli"):
            self.meshtastic_devices[from_id]["last_lat"] = pb.pli.latitude_i * 0.0000001
            self.meshtastic_devices[from_id]["last_lon"] = pb.pli.longitude_i * 0.0000001
            self.meshtastic_devices[from_id]["last_alt"] = pb.pli.altitude
            self.meshtastic_devices[from_id]["course"] = pb.pli.course
            self.meshtastic_devices[from_id]["speed"] = pb.pli.speed
            return self.cot(pb, from_id, to_id, portnum)
        elif pb.HasField("chat"):
            self.logger.debug(
                "Got chat: {} {}->{}: {}".format(
                    unishox2.decompress(pb.chat.to, len(pb.chat.to)),
                    from_id,
                    to_id,
                    unishox2.decompress(pb.chat.message, len(pb.chat.message)),
                )
            )

            chatroom = unishox2.decompress(pb.chat.to, len(pb.chat.to))
            message_uid = str(uuid.uuid4())

            from_uid = sender_callsign = from_id
            if from_uid in self.meshtastic_devices:
                from_uid = self.meshtastic_devices[from_id]["uid"]
                sender_callsign = self.meshtastic_devices[from_id]["long_name"]

            uid = "GeoChat.{}.{}.{}".format(from_uid, chatroom, message_uid)

            event, detail = self.cot(
                pb, from_uid, to_id, portnum, how="h-g-i-g-o", cot_type="b-t-f", uid=uid
            )

            chat = SubElement(
                detail,
                "__chat",
                {
                    "chatroom": "All Chat Rooms",
                    "groupOwner": "false",
                    "id": chatroom,
                    "messageId": message_uid,
                    "parent": "RootContactGroup",
                    "senderCallsign": sender_callsign,
                },
            )
            SubElement(chat, "chatgrp", {"id": chatroom, "uid0": from_uid, "uid1": chatroom})
            SubElement(detail, "link", {"relation": "p-p", "type": "a-f-G-U-C", "uid": from_uid})
            remarks = SubElement(
                detail,
                "remarks",
                {
                    "source": "BAO.F.ATAK.{}".format(from_uid),
                    "time": datetime.datetime.now(datetime.timezone.utc).strftime(
                        "%Y-%m-%dT%H:%M:%SZ"
                    ),
                    "to": chatroom,
                },
            )
            remarks.text = unishox2.decompress(pb.chat.message, len(pb.chat.message))

            return event

    def protobuf_to_cot(self, pb, from_id, to_id, portnum, meshtastic_id):
        self.logger.debug(from_id + " " + to_id + " " + portnum + " " + meshtastic_id)
        event = None

        if from_id not in self.meshtastic_devices:
            self.meshtastic_devices[from_id] = {
                "hw_model": "",
                "long_name": f"Meshtastic {str(from_id)[-4:]}",
                "short_name": str(from_id)[-4:],
                "macaddr": "",
                "firmware_version": "",
                "last_lat": "0.0",
                "last_lon": "0.0",
                "battery": 0,
                "meshtastic_id": meshtastic_id,
                "voltage": 0,
                "uptime": 0,
                "last_alt": "9999999.0",
                "course": "0.0",
                "speed": "0.0",
                "team": "Cyan",
                "role": "Team Member",
                "uid": None,
            }

        if portnum == "MAP_REPORT_APP" or (portnum == "POSITION_APP" and pb.latitude_i):
            event = self.position(pb, from_id, to_id, portnum)
        elif portnum == "NODEINFO_APP":
            event = self.node_info(pb, from_id, to_id, portnum)
        elif portnum == "TEXT_MESSAGE_APP":
            event = self.text_message(pb, from_id, to_id, portnum)
        elif portnum == "ATAK_PLUGIN":
            event = self.atak_plugin(pb, from_id, to_id, portnum)
        elif portnum == "TELEMETRY_APP":
            self.telemetry(pb, from_id, to_id, portnum)

        try:
            if event:
                uid = self.meshtastic_devices[from_id]["uid"]
                if not uid:
                    uid = from_id
                message = json.dumps({"uid": uid, "cot": tostring(event).decode("utf-8")})

                self.rabbit_channel.basic_publish(
                    exchange="firehose",
                    routing_key="",
                    body=message,
                    properties=pika.BasicProperties(
                        expiration=self.context.app.config.get("OTS_RABBITMQ_TTL")
                    ),
                )

                if portnum == "TEXT_MESSAGE_APP":
                    try:
                        if to_id == "all":
                            self.rabbit_channel.basic_publish(
                                exchange="groups",
                                routing_key=f"{self.context.app.config.get('OTS_MESHTASTIC_GROUP')}.OUT",
                                body=message,
                                properties=pika.BasicProperties(
                                    expiration=self.context.app.config.get("OTS_RABBITMQ_TTL")
                                ),
                            )
                        else:
                            for meshtastic_device in self.meshtastic_devices:
                                meshtastic_device = self.meshtastic_devices[meshtastic_device]
                                if meshtastic_device["meshtastic_id"] == to_id:
                                    self.rabbit_channel.basic_publish(
                                        exchange="dms",
                                        routing_key=meshtastic_device["uid"],
                                        body=message,
                                        properties=pika.BasicProperties(
                                            expiration=self.context.app.config.get(
                                                "OTS_RABBITMQ_TTL"
                                            )
                                        ),
                                    )
                    except BaseException as e:
                        self.logger.error("Failed to publish chat message: {}".format(e))
                elif portnum == "ATAK_PLUGIN" and pb.HasField("chat"):
                    try:
                        to = unishox2.decompress(pb.chat.to, len(pb.chat.to))
                        if to in self.meshtastic_devices:
                            self.rabbit_channel.basic_publish(
                                exchange="dms", routing_key=to, body=message
                            )
                        else:
                            self.rabbit_channel.basic_publish(
                                exchange="groups",
                                routing_key=f"{self.context.app.config.get('OTS_MESHTASTIC_GROUP')}.OUT",
                                body=message,
                                properties=pika.BasicProperties(
                                    expiration=self.context.app.config.get("OTS_RABBITMQ_TTL")
                                ),
                            )
                    except BaseException as e:
                        self.logger.error(
                            "Failed to publish chat message to {}: {}".format(
                                unishox2.decompress(pb.chat.to, len(pb.chat.to)), e
                            )
                        )
                        self.logger.error(traceback.format_exc())
                else:
                    self.rabbit_channel.basic_publish(
                        exchange="groups",
                        routing_key=f"{self.context.app.config.get('OTS_MESHTASTIC_GROUP')}.OUT",
                        body=message,
                        properties=pika.BasicProperties(
                            expiration=self.context.app.config.get("OTS_RABBITMQ_TTL")
                        ),
                    )
        except BaseException as e:
            self.logger.error(str(e))
            self.logger.error(traceback.format_exc())



"""Removed unique constraint on ATAK plugin names

Revision ID: eaf2af6f6b22
Revises: 8d6f811f4fd7
Create Date: 2025-11-08 16:28:41.700279

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "eaf2af6f6b22"
down_revision = "8d6f811f4fd7"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        with op.batch_alter_table("packages", schema=None) as batch_op:
            batch_op.drop_constraint(batch_op.f("packages_package_name_key"), type_="unique")
    except BaseException as e:
        print(e)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("packages", schema=None) as batch_op:
        batch_op.create_unique_constraint(batch_op.f("packages_package_name_key"), ["package_name"])

    # ### end Alembic commands ###



import pathlib
from dataclasses import dataclass
from urllib.parse import urlparse

from flask import current_app as app
from flask import request
from sqlalchemy import Boolean, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db
from opentakserver.functions import iso8601_string_from_datetime


@dataclass
class VideoRecording(db.Model):
    __tablename__ = "video_recordings"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    segment_path: Mapped[str] = mapped_column(String(255), unique=True)
    path: Mapped[str] = mapped_column(String(255), ForeignKey("video_streams.path"))
    in_progress: Mapped[bool] = mapped_column(Boolean)
    start_time: Mapped[DateTime] = mapped_column(DateTime, nullable=True)
    stop_time: Mapped[DateTime] = mapped_column(DateTime, nullable=True)
    duration: Mapped[int] = mapped_column(Integer, nullable=True)
    width: Mapped[int] = mapped_column(Integer, nullable=True)
    height: Mapped[int] = mapped_column(Integer, nullable=True)
    video_codec: Mapped[str] = mapped_column(String(255), nullable=True)
    video_bitrate: Mapped[int] = mapped_column(Integer, nullable=True)
    audio_codec: Mapped[str] = mapped_column(String(255), nullable=True)
    audio_bitrate: Mapped[int] = mapped_column(Integer, nullable=True)
    audio_samplerate: Mapped[int] = mapped_column(Integer, nullable=True)
    audio_channels: Mapped[int] = mapped_column(Integer, nullable=True)
    file_size: Mapped[int] = mapped_column(Integer, nullable=True)
    video_stream = relationship("VideoStream", back_populates="recordings")

    def serialize(self):
        return {
            "segment_path": self.segment_path,
            "path": self.path,
            "in_progress": self.in_progress,
        }

    def to_json(self):
        with app.app_context():
            url = urlparse(request.url_root)
            protocol = url.scheme
            hostname = url.hostname
            port = url.port
            if not port and protocol == "https":
                port = 443
            elif not port and protocol == "http":
                port = 80

            return {
                "id": self.id,
                "segment_path": self.segment_path,
                "path": self.path,
                "in_progress": self.in_progress,
                "start_time": (
                    iso8601_string_from_datetime(self.start_time) if self.start_time else None
                ),
                "stop_time": (
                    iso8601_string_from_datetime(self.stop_time) if self.stop_time else None
                ),
                "duration": self.duration,
                "filename": pathlib.Path(self.segment_path).name,
                "width": self.width,
                "height": self.height,
                "video_codec": self.video_codec,
                "video_bitrate": self.video_bitrate,
                "audio_codec": self.audio_codec,
                "audio_bitrate": self.audio_bitrate,
                "audio_samplerate": self.audio_samplerate,
                "audio_channels": self.audio_channels,
                "file_size": self.file_size,
                "thumbnail": f"{protocol}://{hostname}:{port}/api/videos/thumbnail?path={self.path}&recording={pathlib.Path(self.segment_path).name}",
            }



"""Added CasEvac fields

Revision ID: 3a7ea7ead17b
Revises: d68964dab66b
Create Date: 2025-10-03 04:02:44.032178

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3a7ea7ead17b"
down_revision = "d68964dab66b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("casevac", schema=None) as batch_op:
        batch_op.add_column(sa.Column("convenience", sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column("urgent_surgical", sa.String(length=255), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("casevac", schema=None) as batch_op:
        batch_op.drop_column("urgent_surgical")
        batch_op.drop_column("convenience")

    # ### end Alembic commands ###



import semver
from flask_wtf import FlaskForm
from flask_wtf.file import FileAllowed, FileRequired
from wtforms import BooleanField, FileField, StringField
from wtforms.validators import DataRequired, Optional

from opentakserver.functions import false_values


class PackageForm(FlaskForm):
    platform = StringField(default="Android")
    plugin_type = StringField(default="plugin")
    apk = FileField(validators=[FileRequired(), FileAllowed(["apk"])])
    icon = FileField(validators=[Optional()])
    description = StringField(validators=[Optional()])
    install_on_enrollment = BooleanField(false_values=false_values)
    install_on_connection = BooleanField(false_values=false_values)
    atak_version = StringField(validators=[Optional()])


class PackageUpdateForm(FlaskForm):
    package_name = StringField(validators=[DataRequired()])
    install_on_enrollment = BooleanField(false_values=false_values)
    install_on_connection = BooleanField(false_values=false_values)
    atak_version = StringField(validators=[Optional()])

    @staticmethod
    def validate_atak_version(form, field):
        # Plugins for ATAK 5.4.0 and older won't have an atak_version
        if field.data:
            # semver.Version.parse() will raise ValueError if it's not a semver string.
            # Flask-wtf will mark the data as invalid when ValueError is raised
            semver.Version.parse(field.data)



import datetime
from dataclasses import dataclass

from flask import current_app as app
from flask import request
from sqlalchemy import JSON, Boolean, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db, logger
from opentakserver.functions import iso8601_string_from_datetime
from opentakserver.models.MissionRole import MissionRole


@dataclass
class Mission(db.Model):
    __tablename__ = "missions"

    INVITE = "INVITE"
    DELETE = "DELETE"
    CREATE = "CREATE"

    name: Mapped[str] = mapped_column(String(255), primary_key=True)
    description: Mapped[str] = mapped_column(String(255), nullable=True)
    chat_room: Mapped[str] = mapped_column(String(255), nullable=True)
    base_layer: Mapped[str] = mapped_column(String(255), nullable=True)
    bbox: Mapped[str] = mapped_column(String(255), nullable=True)
    path: Mapped[str] = mapped_column(String(255), nullable=True)
    classification: Mapped[str] = mapped_column(String(255), nullable=True)
    tool: Mapped[str] = mapped_column(String(255), nullable=True)
    group: Mapped[str] = mapped_column(String(255), nullable=True)
    default_role: Mapped[str] = mapped_column(String(255), nullable=True)
    keywords: Mapped[JSON] = mapped_column(JSON, nullable=True)
    creator_uid: Mapped[str] = mapped_column(
        String(255), ForeignKey("euds.uid", ondelete="CASCADE"), nullable=True
    )
    create_time: Mapped[datetime] = mapped_column(DateTime, nullable=True)
    external_data: Mapped[JSON] = mapped_column(JSON, nullable=True)
    feeds: Mapped[JSON] = mapped_column(JSON, nullable=True)
    map_layers: Mapped[JSON] = mapped_column(JSON, nullable=True)
    invite_only: Mapped[bool] = mapped_column(Boolean, nullable=True)
    expiration: Mapped[int] = mapped_column(Integer, nullable=False, default=-1)
    guid: Mapped[str] = mapped_column(String(255), nullable=True)
    password_protected: Mapped[bool] = mapped_column(Boolean, nullable=True)
    password: Mapped[str] = mapped_column(String(255), nullable=True)
    invitations = relationship(
        "MissionInvitation", cascade="all, delete-orphan", back_populates="mission"
    )
    roles = relationship("MissionRole", cascade="all, delete-orphan", back_populates="mission")
    mission_changes = relationship(
        "MissionChange", cascade="all, delete-orphan", back_populates="mission", uselist=True
    )
    contents = relationship(
        "MissionContent",
        cascade="all, delete",
        secondary="mission_content_mission",
        back_populates="mission",
        uselist=True,
    )
    cots = relationship("CoT", back_populates="mission", uselist=True)
    uids = relationship("MissionUID", cascade="all, delete-orphan", back_populates="mission")
    mission_logs = relationship(
        "MissionLogEntry", cascade="all, delete-orphan", back_populates="mission"
    )
    owner = relationship("EUD", back_populates="owned_missions")
    groups = relationship(
        "Group",
        secondary="groups_missions",
        viewonly=True,
        back_populates="missions",
        cascade="all, delete",
    )

    def serialize(self):
        return {
            "description": self.description,
            "chat_room": self.chat_room,
            "base_layer": self.base_layer,
            "bbox": self.bbox,
            "path": self.path,
            "classification": self.classification,
            "tool": self.tool,
            "group": self.group or "__ANON__",
            "default_role": self.default_role,
            "keywords": self.keywords,
            "creator_uid": self.creator_uid,
            "create_time": self.create_time,
            "external_data": self.external_data,
            "feeds": self.feeds,
            "map_layers": self.map_layers,
            "invite_only": self.invite_only,
            "expiration": self.expiration,
            "guid": self.guid,
            "password_protected": self.password_protected,
            "password": self.password,
        }

    def to_json(self):
        url = request.url_root.replace("http://", "").replace("https://", "").replace("/", "")
        json = {
            "name": self.name,
            "description": self.description or "",
            "chatRoom": self.chat_room or "",
            "baseLayer": self.base_layer or "",
            "bbox": self.bbox or "",
            "path": self.path or "",
            "classification": self.classification or "",
            "tool": self.tool or "public",
            "defaultRole": self.default_role or "",
            "keywords": self.keywords if self.keywords else [],
            "creatorUid": self.creator_uid or "",
            "createTime": iso8601_string_from_datetime(self.create_time) or "",
            "externalData": self.external_data if self.external_data else [],
            "feeds": self.feeds if self.feeds else [],
            "mapLayers": self.map_layers if self.map_layers else [],
            "inviteOnly": self.invite_only if self.invite_only is not None else False,
            "expiration": self.expiration if self.expiration is not None else -1,
            "guid": self.guid or "",
            "uids": [uid.to_json() for uid in self.uids],
            "contents": [content.to_json() for content in self.contents],
            "passwordProtected": (
                self.password_protected if self.password_protected is not None else False
            ),
            "missionChanges": [mission_change.to_json() for mission_change in self.mission_changes],
            "qr_code": f"{url}:{app.config.get('OTS_SSL_STREAMING_PORT')}:ssl,{url}-{app.config.get('OTS_MARTI_HTTPS_PORT')}-ssl-{self.name},{self.name}",
            "owner": self.owner.to_json() if self.owner else None,
            "groups": [],
        }

        for group in self.groups:
            json["groups"].append({"id": group.id, "name": group.name})

        if self.default_role == MissionRole.MISSION_SUBSCRIBER or not self.default_role:
            json["defaultRole"] = MissionRole.SUBSCRIBER_ROLE

        elif self.default_role == MissionRole.MISSION_OWNER:
            json["defaultRole"] = MissionRole.OWNER_ROLE

        else:
            json["defaultRole"] = MissionRole.READ_ONLY_ROLE

        return json

    def to_marti_json(self):
        return_value = self.to_json()

        return_value["groups"] = []
        for group in self.groups:
            return_value["groups"].append(group.name)

        return return_value



import os
import pathlib
import traceback

import bleach
from flask import Blueprint
from flask import current_app as app
from flask import jsonify, request, send_from_directory
from flask_babel import gettext
from flask_security import auth_required

from opentakserver.blueprints.ots_api.api import paginate, search
from opentakserver.extensions import db, logger
from opentakserver.models.VideoRecording import VideoRecording
from opentakserver.models.VideoStream import VideoStream

video_api_blueprint = Blueprint("video_api_blueprint", __name__)


@video_api_blueprint.route("/api/videos/thumbnail", methods=["GET"])
@auth_required()
def thumbnail():
    path = request.args.get("path")
    recording = request.args.get("recording")
    if not path:
        return jsonify({"success": False, "error": "Please specify a path"}), 400

    if recording and os.path.exists(
        os.path.join(
            app.config.get("OTS_DATA_FOLDER"), "mediamtx", "recordings", path, recording + ".png"
        )
    ):
        return send_from_directory(
            os.path.join(app.config.get("OTS_DATA_FOLDER"), "mediamtx", "recordings", path),
            recording + ".png",
        )

    elif os.path.exists(
        os.path.join(app.config.get("OTS_DATA_FOLDER"), "mediamtx", "recordings", path)
    ):
        return send_from_directory(
            os.path.join(app.config.get("OTS_DATA_FOLDER"), "mediamtx", "recordings", path),
            "thumbnail.png",
        )

    return jsonify({"success": False, "error": gettext("Please specify a valid path")}), 400


@video_api_blueprint.route("/api/videos/recordings")
@auth_required()
def video_recordings():
    query = db.session.query(VideoRecording)
    query = search(query, VideoRecording, "path")

    return paginate(query, VideoRecording)


@video_api_blueprint.route("/api/videos/recording", methods=["GET", "DELETE", "HEAD"])
@auth_required()
def download_recording():
    if not request.args.get("id"):
        return jsonify({"success": False, "error": "Please specify a recording ID"}), 400
    recording_id = bleach.clean(request.args.get("id"))

    if recording_id.isdigit():
        recording_id = int(recording_id)
    else:
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext(
                        "Invalid recording_id: %(recording_id)s", recording_id=recording_id
                    ),
                }
            ),
            400,
        )

    try:
        recording = db.session.execute(
            db.session.query(VideoRecording).filter(VideoRecording.id == recording_id)
        ).first()[0]

        if request.method == "GET":
            filename = pathlib.Path(recording.segment_path)
            return send_from_directory(filename.parent, filename.name)
        elif request.method == "DELETE":
            os.remove(recording.segment_path)
            db.session.delete(recording)
            db.session.commit()
            return jsonify({"success": True})
        elif request.method == "HEAD":
            return "", 200
    except:
        logger.error(traceback.format_exc())
        return jsonify({"success": False, "error": gettext("Recording not found")}), 404


@video_api_blueprint.route("/api/video_streams")
@auth_required()
def get_video_streams():
    query = db.session.query(VideoStream)
    query = search(query, VideoStream, "username")
    query = search(query, VideoStream, "protocol")
    query = search(query, VideoStream, "path")
    query = search(query, VideoStream, "uid")

    return paginate(query, VideoStream)



import base64
import os
import traceback

import bleach
import sqlalchemy
from flask import Blueprint
from flask import current_app as app
from flask import jsonify, request
from flask_babel import gettext
from flask_security import auth_required
from meshtastic import apponly_pb2, channel_pb2

from opentakserver.blueprints.ots_api.api import paginate, search
from opentakserver.extensions import db, logger
from opentakserver.models.Meshtastic import MeshtasticChannel

meshtastic_api_blueprint = Blueprint("meshtastic_api_blueprint", __name__)


@meshtastic_api_blueprint.route("/api/meshtastic/channel", methods=["POST"])
@auth_required()
def create_channel():
    channel_set = apponly_pb2.ChannelSet()
    channel_settings = channel_pb2.ChannelSettings()

    # Parse the settings from a Meshtastic URL
    if "url" in request.json.keys() and request.json.get("url"):
        try:
            settings = base64.urlsafe_b64decode(
                bleach.clean(request.json.get("url").split("#")[-1]) + "=="
            )
            channel_set.ParseFromString(settings)
            url = request.json.get("url")

        except BaseException as e:
            logger.error("Failed to parse Meshtastic  URL: {}".format(e))
            logger.error(traceback.format_exc())
            return jsonify({"success": False, "error": str(e)}), 400

    # Make the protobuf from settings in order to make the URL
    else:
        try:
            # channel_set.lora_config.use_preset = 'modem_preset' in request.json.keys()
            channel_set.lora_config.modem_preset = (
                bleach.clean(request.json.get("modem_preset"))
                if request.json.get("modem_preset")
                else None
            )
            channel_set.lora_config.region = (
                bleach.clean(request.json.get("lora_region"))
                if request.json.get("lora_region")
                else None
            )
            channel_set.lora_config.hop_limit = (
                request.json.get("lora_hop_limit") if request.json.get("lora_hop_limit") else None
            )
            channel_set.lora_config.tx_enabled = request.json.get("lora_tx_enabled")
            channel_set.lora_config.tx_power = (
                request.json.get("lora_tx_power") if request.json.get("lora_tx_power") else None
            )
            channel_set.lora_config.sx126x_rx_boosted_gain = request.json.get(
                "lora_sx126x_rx_boosted_gain"
            )
            channel_set.lora_config.use_preset = True

            if "psk" in request.json.keys() and request.json.get("psk"):
                channel_settings.psk = base64.urlsafe_b64decode(
                    bleach.clean(request.json.get("psk"))
                )
            else:
                # Zero bytes indicates no encryption
                channel_settings.psk = bytes(0)
            channel_settings.name = (
                bleach.clean(request.json.get("name")) if request.json.get("name") else None
            )
            channel_settings.uplink_enabled = request.json.get("uplink_enabled")
            channel_settings.downlink_enabled = request.json.get("downlink_enabled")
            channel_settings.module_settings.position_precision = (
                request.json.get("position_precision")
                if request.json.get("position_precision")
                else None
            )

            channel_set.settings.append(channel_settings)

        except BaseException as e:
            logger.error("Failed to save Meshtastic channel: {}".format(e))
            logger.error(traceback.format_exc())
            return jsonify({"success": False, "error": str(e)}), 400

    return save_channel(channel_set)


def save_channel(channel_set: apponly_pb2.ChannelSet):
    for channel_settings in channel_set.settings:
        # Make a unique URL for each channel in a channel set
        single_channel_set = apponly_pb2.ChannelSet()
        single_channel_set.lora_config.CopyFrom(channel_set.lora_config)
        single_channel_set.settings.append(channel_settings)
        url = "https://meshtastic.org/e/#" + base64.urlsafe_b64encode(
            single_channel_set.SerializeToString()
        ).decode("utf-8")

        if channel_settings.downlink_enabled and channel_settings.name not in app.config.get(
            "OTS_MESHTASTIC_DOWNLINK_CHANNELS"
        ):
            app.config.get("OTS_MESHTASTIC_DOWNLINK_CHANNELS")[
                channel_settings.name
            ] = channel_settings.psk
            logger.debug("Added {} to channels".format(channel_settings.name))

        meshtastic_channel_settings = MeshtasticChannel()
        meshtastic_channel_settings.psk = base64.urlsafe_b64encode(channel_settings.psk).decode(
            "ascii"
        )
        meshtastic_channel_settings.name = channel_settings.name
        meshtastic_channel_settings.uplink_enabled = channel_settings.uplink_enabled
        meshtastic_channel_settings.downlink_enabled = channel_settings.downlink_enabled
        meshtastic_channel_settings.position_precision = (
            channel_settings.module_settings.position_precision
        )
        meshtastic_channel_settings.lora_region = channel_set.lora_config.region
        meshtastic_channel_settings.lora_hop_limit = channel_set.lora_config.hop_limit
        meshtastic_channel_settings.lora_tx_enabled = channel_set.lora_config.tx_enabled
        meshtastic_channel_settings.lora_tx_power = channel_set.lora_config.tx_power
        meshtastic_channel_settings.lora_sx126x_rx_boosted_gain = (
            channel_set.lora_config.sx126x_rx_boosted_gain
        )
        meshtastic_channel_settings.modem_preset = channel_set.lora_config.modem_preset
        meshtastic_channel_settings.url = url

        try:
            db.session.add(meshtastic_channel_settings)
            db.session.commit()
        except sqlalchemy.exc.IntegrityError:
            logger.error("Channel with URL {} already exists".format(url))
            # return jsonify({'success': False, 'error': "Channel already exists"}), 400
            continue
        except BaseException as e:
            logger.error("Failed to save Meshtastic channel: {}".format(e))
            logger.error(traceback.format_exc())
            return jsonify({"success": False, "error": str(e)}), 400

    return jsonify({"success": True, "url": url})


@meshtastic_api_blueprint.route("/api/meshtastic/channel", methods=["GET"])
@auth_required()
def get_channel():
    query = db.session.query(MeshtasticChannel)
    query = search(query, MeshtasticChannel, "id")
    query = search(query, MeshtasticChannel, "name")
    query = search(query, MeshtasticChannel, "url")

    return paginate(query, MeshtasticChannel)


@meshtastic_api_blueprint.route("/api/meshtastic/channel", methods=["DELETE"])
@auth_required()
def delete_channel():
    if "url" not in request.args.keys():
        return (
            jsonify(
                {
                    "success": False,
                    "error": gettext("Please provide the URL of the channel to delete"),
                }
            ),
            400,
        )

    url = bleach.clean(request.args.get("url"))

    try:
        channel = db.session.query(MeshtasticChannel).filter(MeshtasticChannel.url == url).first()
        db.session.delete(channel)
        db.session.commit()

        if channel.name in app.config.get("OTS_MESHTASTIC_DOWNLINK_CHANNELS"):
            app.config.get("OTS_MESHTASTIC_DOWNLINK_CHANNELS").pop(channel.name)

    except BaseException as e:
        logger.error("Failed to delete Meshtastic channel: {}".format(e))
        logger.error(traceback.format_exc())
        return jsonify({"success": False, "error": str(e)}), 400

    return jsonify({"success": True})


@meshtastic_api_blueprint.route("/api/meshtastic/generate_psk")
@auth_required()
def generate_psk():
    return (
        jsonify({"success": True, "psk": base64.urlsafe_b64encode(os.urandom(32)).decode("ascii")}),
        200,
    )
